import pandas as pd
import numpy as np

from sklearn.utils.metaestimators import _safe_split
import sklearn.preprocessing as preprocessing
from sklearn.base import BaseEstimator, TransformerMixin

from pyadlml.dataset._representations.raw import create_raw, resample_raw
from pyadlml.dataset._representations.changepoint import create_changepoint, resample_changepoint
from pyadlml.dataset._representations.lastfired import create_lastfired, resample_last_fired
from pyadlml.dataset._core._dataset import label_data
from pyadlml.constants import ACTIVITY, TIME, DEVICE, VALUE, END_TIME, START_TIME, \
                              ENC_RAW, ENC_LF, ENC_CP, REPS
from pyadlml.dataset._core.devices import create_device_info_dict
from pyadlml.pipeline import XOrYTransformer, XAndYTransformer, YTransformer


class Df2Numpy(TransformerMixin, XOrYTransformer):
    """ Transforms dataframes to numpy arrays by respecting
    the column order
    """
    _x_columns = []
    _y_columns = []

    def __init__(self):
        pass

    def fit(self, X, y=None):
        self._x_columns = X.columns
        if y is not None:
            self._y_columns = y.columns


    def fit_transform(self, X, y):
        self.fit(X,y)
        return self.transform(X,y)

    def transform(self, X, y):
        # reorder columns
        X = X[self._x_columns]
        y = y[self._y_columns]

        #
        X_np = X.values
        y_np = y.values
        return X_np, y_np


class DfCaster(TransformerMixin, XOrYTransformer):
    """ Transforms dataframes to torch tensors by respecting
    the column order
    """
    _x_columns = []
    _y_columns = []
    Df2Numpy = 'df->np'
    Df2Torch = 'df->torch'

    def __init__(self, x_conv, y_conv):
        self.y_conv = y_conv
        self.x_conv = x_conv

    def fit(self, X, y=None):
        assert self.y_conv in [self.Df2Numpy, self.Df2Torch]
        assert self.x_conv in [self.Df2Numpy, self.Df2Torch]

        self._x_columns = X.columns
        if y is not None:
            self._y_columns = y.columns


    def fit_transform(self, X, y=None):
        self.fit(X, y=y)
        return self.transform(X, y=y)

    def transform(self, X, y=None):
        """
        Parameters
        ----------
        X : pd
        """

        # reorder columns
        X = X[self._x_columns]

        # cast to respective format
        if self.x_conv == self.Df2Torch:
            import torch
            X_prime = torch.tensor(X.values, dtype=torch.float32)
        elif self.x_conv == self.Df2Numpy:
            X_prime = X.values.astype('float32')

        if y is not None:
            y = y[self._y_columns]
            if self.y_conv == self.Df2Torch:
                import torch
                y_prime = torch.tensor(y.values, dtype=torch.int64)
            elif self.y_conv == self.Df2Numpy:
                # ensure that y has dim (N,) and is of correct type
                y_prime = y.values.squeeze().astype('int64')
            return X_prime, y_prime
        else:
            return X_prime


class Df2Torch(TransformerMixin, XOrYTransformer):
    """ Transforms dataframes to torch tensors by respecting
    the column order
    """
    _x_columns = []
    _y_columns = []

    def __init__(self, only_y=False, only_X=False):
        self.only_y = only_y
        self.only_X = only_X

    def fit(self, X, y=None):
        self._x_columns = X.columns
        if y is not None:
            self._y_columns = y.columns


    def fit_transform(self, X, y):
        self.fit(X,y)
        return self.transform(X,y)

    def transform(self, X, y):
        """
        Parameters
        ----------
        X : pd
        """
        import torch

        # reorder columns
        X = X[self._x_columns]
        y = y[self._y_columns]

        # cast to tensors
        y_tensor = torch.tensor(y.values, dtype=torch.int64)
        X_tensor = torch.tensor(X.values, dtype=torch.float32)

        return X_tensor, y_tensor


class IdentityTransformer(TransformerMixin, XOrYTransformer):
    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        return X, y


class RandomUnderSampler(TransformerMixin, XAndYTransformer):

    def __init__(self, sampling_strategy):
        self.sampling_strategy = sampling_strategy

    def fit(self, X, y, *args):
        from imblearn.under_sampling import RandomUnderSampler as ImbRUS
        self.rus_ = ImbRUS(sampling_strategy=self.sampling_strategy)
        return self

    def fit_transform(self, X, y, **fit_params):
        self.fit(X, y)
        return self.transform(X, y)

    def transform(self, X, y):
        return self.rus_.fit_resample(X,y)


class DropNanRows(XOrYTransformer):
    """
    Drops all rows where the label or the data has a nan-value
    """
    def fit(self, X,y=None):
        return self

    def fit_transform(self, X, y=None):
        return self.transform(X, y=y)

    def transform(self, X, y=None):
        """
        If X is given transform y
        """
        if y is not None:
            if isinstance(y, pd.DataFrame):
                nan_mask = y[ACTIVITY].isna().values
                return X[~nan_mask], y[~nan_mask]

            elif isinstance(y, np.ndarray):
                raise NotImplementedError
        else:
            nan_mask = X.isna().values
            return X[~nan_mask]


class DropDuplicates(TransformerMixin, XOrYTransformer):
    def __init__(self):
        BaseEstimator.__init__(self)
        XOrYTransformer.__init__(self)

    def fit(self, X, y=None):
        return self

    def fit_transform(self, X, y=None, **fit_params):
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        """  Depending if X or y is given drop either all Nans in y or all nans in X or
        all Nans in the union of X and y.

        """
        if y is not None and X is not None:
            assert isinstance(y, pd.DataFrame)
            tmp = X.copy()
            tmp['y'] = y
            tmp = tmp.drop_duplicates()
            y = tmp['y']
            X = tmp.drop('y', axis=1)
        if y is not None and X is None:
            y = y.drop_duplicates()
        if X is not None:
            X = X.drop_duplicates()
        return X, y


class DropTimeIndex(TransformerMixin, XOrYTransformer):
    def __init__(self):
        XOrYTransformer.__init__(self)

    def fit_transform(self, X, y=None, **fit_params):
        """
        drops columns that are not time dependent
        """
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        if X is not None:
            X = X.loc[:, X.columns != TIME]
        if y is not None:
            y = y.loc[:, y.columns != TIME]
        return X, y


class CVSubset(TransformerMixin, XOrYTransformer):
    """ Selects the subset from the whole dataset based on a given data_range.

    Attributes
    ----------
    data_range  : list of either indices or interval
        Either
    y : boolean, default=False
        Only

    """
    def __init__(self, data_range=None, y=False):
        self.data_range = data_range
        XOrYTransformer.__init__(self)

    def set_range(self, data_range):
        self.data_range = data_range

    def _extract_first_sample(self, data_range):
        try:
            sample = data_range[0][0]
        except TypeError:
            sample = data_range[0]
        except IndexError:
            sample = data_range[0]
        return sample


    def fit(self, X, y=None):
        error_msg = "Data range has to be set to either time intervals or list of indicies. Not 'None'"
        assert self.data_range is not None, error_msg

        # determine whether the splits are made based on index ranges or timestamp tuples
        sample = self._extract_first_sample(self.data_range)
        self.timestamps_ = not (isinstance(sample, int) or isinstance(sample, np.int64))

        return self

    def fit_transform(self, X, y=None):
        self.fit(X, y)
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X, y=None):
        """ Selects a subset of X and y based on the data_range and the split type
        """
        if self.timestamps_:
            x_mask = _create_mask(X, self.data_range)
            X = X[x_mask]
            if y is not None:
                assert len(X) == len(y)
                y_mask = _create_mask(y, self.data_range)
                y = y[y_mask]
            return X, y
        else:
            X, y = _safe_split(None, X, y, self.data_range)
            return X, y


def _create_mask(X, date_range):
    x_mask = (X[TIME] == 'false_init')
    if isinstance(date_range, tuple):
        date_range = [date_range]
    for time_pair in date_range:
        x_mask = x_mask | ((time_pair[0] < X[TIME]) & (X[TIME] < time_pair[1]))
    return x_mask


class StateVectorEncoder(BaseEstimator, TransformerMixin):
    """
    Create a State-vector sequence from a device event stream.
    Read more in the :ref:`User Guide <preprocessing_discretization>`.

    Parameters
    ----------
    encode : {'raw', 'changepoint', 'last_fired'}, default='raw'
        Determines the state-vectors encoding.

        raw
            Encode the event stream as vector, where each field
            represents a device. Each entry represents the current
            state that the device is in.

        changepoint
            Encode the event stream as one-hot binary vector, where
            each field represents a device. A one indicates the
            device that produced the event.

        last_fired
            Encode the event stream as one-hot binary vector,
            where each field represents a device. The device that produced
            the last event is indicated with one.

    dt : str, optional, default=None
        The timeslices resolution for discretizing the event stream. If
        set to None the event stream is not discretized.

    Attributes
    ----------
    encode : string
        The encoding for the vectors
    dt : str or None
        Determines the binsize that the event stream is discretized into
    data_info_ : dict or None
        Dictionary containing further device information such as most likely value
        or device datatype. This is used to speed up internal computation.
    classes_ : list
        A list of all devices. The lists order corresponds to the device order of the
        transformed state-vectors.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_amsterdam
    >>> data = fetch_amsterdam()
    >>> from pyadlml.preprocessing import StateVectorEncoder
    >>> sve = StateVectorEncoder(encode='raw')
    >>> X = sve.fit_transform(data.df_devices)
    >>> sve2 = StateVectorEncoder(encode='changepoint+raw', dt='6s')
    >>> X_upsampled = sve2.fit_transform(data.df_devices)

    """

    def __init__(self, encode=ENC_RAW, dt=None):
        self.encode = encode
        self.dt = dt
        self.data_info_ = None
        self.classes_ = []

    @classmethod
    def _is_valid_encoding(cls, encoding):
        from itertools import chain, permutations
        def helper(iterable):
            s = list(iterable)
            return chain.from_iterable(permutations(s, r) for r in range(len(s)+1))
        # generate all possible valid combinations
        valid_combos = ['+'.join(combo) for combo in helper(REPS)]
        return encoding in valid_combos

    def fit(self, df_devs, y=None):
        """
        Fit the estimator.

        Parameters
        ----------
        df_devs : pd.DataFrame, optional
            recorded devices from a dataset. For more information refer to
            :ref:`user guide<device_dataframe>`.
        y : None
            Ignored. This parameter exists only for compatibility with
            :class:`~sklearn.pipeline.Pipeline`.

        Returns
        -------
        self
        """
        assert StateVectorEncoder._is_valid_encoding(self.encode)

        # create hashmap off all input features with mean for numerical
        # and most common value for binary or categorical features
        self.data_info_ = create_device_info_dict(df_devs)
        self.classes_ = self.data_info_.keys()
        return self

    def transform(self, df_devs=None,y=None, dev_pre_values={}):
        """
        Discretize the data.

        Parameters
        ----------
        df_devs : array-like of shape (n_samples, n_features)
            The device dataframe to be transformed.
        y : None
            todo copy standard scipy sentence
        dev_pre_values : dict
            A dictionary containing a device-value mapping that is used
            to fill in information . If no dictionary is given the values are infered by
            most likely values for categorical and boolean devices.

        Returns
        -------
        Xt : {ndarray, sparse matrix}, dtype={np.float32, np.float64}
            Data in the binned space. Will be a sparse matrix if
            `self.encode='onehot'` and ndarray otherwise.
        """
        PRAEFIX_LF = 'lf_'
        PRAEFIX_CP = 'cp_'


        df_lst = []
        iters = self.encode.split('+')
        for enc in iters:
            if enc == ENC_RAW:
                data = create_raw(df_devs, self.data_info_, dev_pre_values=dev_pre_values)
                if self.dt is not None:
                    data = resample_raw(data, df_dev=df_devs, dt=self.dt,
                                        most_likely_values=self.data_info_
                                        )

                # convert boolean data into integers (1,0)
                dev_bool = [dev for dev in self.data_info_.keys() if self.data_info_[dev]['dtype'] == 'boolean']
                data[dev_bool] = data[dev_bool].astype(int)

            elif enc == ENC_CP:
                data = create_changepoint(df_devs)
                if self.dt is not None:
                    data = resample_changepoint(data, self.dt)

                # set values that are missing in transform but were present when fitting to 0
                dev_diff = set(self.classes_) - set(data.columns)
                if len(dev_diff) > 0:
                    for dev in dev_diff:
                        data[dev] = 0

                # add prefix to make column names unique
                if len(iters) > 1:
                    data.columns = [TIME] + list(map(PRAEFIX_CP.__add__, data.columns[1:]))


            elif enc == ENC_LF:
                data = create_lastfired(df_devs)
                if self.dt is not None:
                    data = resample_last_fired(data, self.dt)

                # set values that are missing in transform but were present when fitting to 0
                dev_diff = set(self.classes_) - set(data.columns)
                if len(dev_diff) > 0:
                    for dev in dev_diff:
                        data[dev] = 0

                # add prefix to make column names unique
                if len(iters) > 1:
                    data.columns = [TIME] + list(map(PRAEFIX_LF.__add__, data.columns[1:]))

            else:
                raise ValueError
            data = data.set_index(TIME)
            df_lst.append(data)

        data = pd.concat(df_lst, axis=1).reset_index()
        return data


    def fit_transform(self, df_devs, y=None):
        """

        Parameters
        ----------
        df_devs : pd.DataFrame, optional
            recorded devices from a dataset. For more information refer to
            :ref:`user guide<device_dataframe>`.
        y : None
            Ignored. This parameter exists only for compatibility with
            :class:`~sklearn.pipeline.Pipeline`.
        """
        self.fit(df_devs)
        return self.transform(df_devs)


class LabelEncoder(BaseEstimator, TransformerMixin, YTransformer):
    """
    Labels a dataframe with corresponding activities

    Attributes
    ----------
    idle : bool, default=False
        If true items that are not

    """
    def __init__(self, idle=False):
        """
        Returns
        -------
        self
        """
        self.idle = idle

    def fit(self, y, X):
        """
        Fit label encoder.

        Parameters
        ----------
        y : pd.DataFrame
            recorded activities from a dataset. Fore more information refer to the
            :ref:`user guide<activity_dataframe>`.

        Returns
        -------
        self : returns an instance of self
        """
        self.lbl_enc_ = preprocessing.LabelEncoder()
        self.df_acts_ = y

        # add actvities to sklearns LabelEncoder
        activities = list(y[ACTIVITY].values)
        if self.idle:
            activities.append('idle')
            # todo code-smells import constant idle activity
        self.lbl_enc_.fit(activities)
        self.classes_ = self.lbl_enc_.classes_
        return self


    def fit_transform(self, y, X):
        """
        Fit label encoder and return encoded labels.

        Parameters
        ----------
        y : pd.DataFrame
            An activity dataframe recorded activities from a dataset. Fore more information refer to the
            :ref:`user guide<activity_dataframe>`.
        X : pd.DataFrame
            A dataframe containing a column with name 'time' of timestamps.

        Returns
        ------
        df : pd.DataFrame
            The X dataframe with an additional column containing the corresponding activity labels
        """
        self.fit(y, X)
        return self.transform(y, X)

    def inverse_transform(self, y, retain_index=False):
        """
        Transform labels back to original encoding.

        Parameters
        ----------
        x : array like or pd.DataFrame or pd.Series
            array of numbers that are transformed to labels
        retain_index : bool, default=False
            TODO

        Returns
        -------
            TODO
        """
        raise NotImplementedError
        #if isinstance(x, np.ndarray):
        #    res = self._lbl_enc.inverse_transform(x)
        #    if retain_index:
        #        return pd.Series(data=res, index=self.df_devs.index[:len(res)])
        #    else:
        #        return res
        #
        #elif isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
        #    tmp_index = x.index
        #    res = self._lbl_enc.inverse_transform(x.values)
        #    return pd.Series(data=res, index=tmp_index)
        #else:
        #    raise ValueError

    def transform(self, y, X=None):
        """

        Parameters
        ----------
        y : pd.DataFrame
            An activity dataframe
        X : pd.DataFrame
            A device dataframe

        """
        # normal case where X an y are provided
        if X is not None:
            df = label_data(X, y, self.idle)
            # remove Nans before encoding the labels and then concatenate again
            if not self.idle:
                nan_mask = df[ACTIVITY].isna()
                df_san_nan = df[~nan_mask]
                encoded_labels = self.lbl_enc_.transform(df_san_nan[ACTIVITY].values)
                new_san_nan = pd.DataFrame(data={TIME: df_san_nan[TIME].values, ACTIVITY: encoded_labels})
                new_nans = pd.DataFrame(data={TIME: df[nan_mask][TIME].values, ACTIVITY: np.nan})
                return pd.concat([new_nans, new_san_nan]).sort_values(by=TIME)
            else:
                encoded_labels = self.lbl_enc_.transform(df[ACTIVITY].values)
                return pd.DataFrame(data={TIME: df[TIME].values, ACTIVITY: encoded_labels})

        ## case where X is a activity dataframe
        #if isinstance(X, pd.DataFrame) and set(X.columns) == {START_TIME, END_TIME, ACTIVITY}:
        #    df = label_data(X, self.df_acts, self.idle)
        #    encoded_labels = self._lbl_enc.fit_transform(df[ACTIVITY].values)
        #    return pd.DataFrame(index=df[TIME], data=encoded_labels, columns=[ACTIVITY])


        ## case where X is a device dataframe
        #if isinstance(X, pd.DataFrame) and TIME in X.columns:
        #    df = label_data(X, self.df_acts_, self.idle)
        #    encoded_labels = self.lbl_enc_.transform(df[ACTIVITY].values)
        #    return pd.DataFrame(data={TIME: df[TIME].values, ACTIVITY: encoded_labels})

        ## return only the labels for a nd array
        #elif isinstance(X, np.ndarray):
        #    return self.lbl_enc_.transform(X)

        else:
            raise ValueError


class SequenceSlicer(TransformerMixin, XOrYTransformer):
    """ Slices a sequence into batches.

    Parameters
    ----------

    .. image:: ../_static/images/many_to_many.svg
       :height: 200px
       :width: 500 px
       :scale: 90%
       :alt: alternate text
       :align: center


    .. image:: ../_static/images/reps/image.svg
       :height: 200px
       :width: 500 px
       :scale: 80%
       :alt: alternate text
       :align: center


    Many-To-Many
    ^^^^^^^^^^^^

    To get *many-to-many* batches use the window size :math:`w` to split the data

    .. math::
        f(X_{N,K},y_{N}) \rightarrow (X_{W, N,K}, y_{W, N})

    In addition you can specify a stride.


    .. code:: python

        from pyadlml.preprocessing import SequenceDicer

        raw = StateVectorEncoder(encode='raw', t_res='10s')\
              .fit_transform(data.df_devices)
        labels = LabelEncoder().fit_transform(raw, data.df_activities)

        X = raw.values
        y = labels.values

        X, y = SequenceDicer(rep='many-to-many', window_length='5m')\
               .fit_transform(X, y)

    Many-To-One
    ^^^^^^^^^^^

    .. math::
        f(X_{N,K},y_{N}) \rightarrow (X_{W, N, K}, y_{N})


    .. code:: python

        from pyadlml.preprocessing import SequenceDicer

        raw = StateVectorEncoder(encode='raw', t_res='10s')\
              .fit_transform(data.df_devices)
        labels = LabelEncoder().fit_transform(raw, data.df_activities)

        X = raw.values
        y = labels.values

        X, y = SequenceDicer(rep='many-to-one', window_length='20s')\
               .fit_transform(X, y)

    """
    REP_M2M = 'many-to-many'
    REP_M2O = 'many-to-one'

    def __init__(self, rep=REP_M2M, window_size=10, stride=1):
        TransformerMixin.__init__(self)
        XOrYTransformer.__init__(self)

        self.rep = rep
        self.window_size = window_size
        self.stride = stride

    def fit(self, X,y=None, **fit_params):
        # TODO not having this method throws an error ! dirty hack
        return self

    def fit_transform(self, X, y=None):
        """
        Parameters
        ----------
        X : nd.array
            Some kind of numpy array
        y : nd.array

        Returns
        -------
        x : nd.array
            todo
        y : nd.array
            todo
        """
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        """

        """
        assert self.rep in [self.REP_M2M, self.REP_M2O]

        if X is not None:
            assert self.window_size < len(X)
            assert self.stride < len(X)

            X = self._transform_X(X)

        if y is not None:
            assert self.window_size < len(y)
            assert self.stride < len(y)

            y = self._transform_Y(y)

        return X, y

    def _calc_new_N(self, n_old):
        return int(np.floor((n_old-self.window_size)/self.stride)+1)

    def _transform_X(self, X):
        """

        """
        n_prime = self._calc_new_N(X.shape[0])
        new_shape =[n_prime, self.window_size, X.shape[1]]

        res = np.zeros(shape=new_shape, dtype='float32')
        if not isinstance(X, np.ndarray):
            import torch
            if isinstance(X, torch.Tensor):
                try:
                    res = torch.zeros(new_shape, dtype=torch.float32)
                except TypeError:
                    print('asdf')
            else:
                raise ValueError('X has to be either ndarray or of type torch tensor')

        for r, i in enumerate(range(0, n_prime*self.stride, self.stride)):
            res[r, :, :] = X[i: i+self.window_size, :]
        return res

    def _transform_Y(self, y):
        """


        """
        dtype = 'int64'
        if self.rep == self.REP_M2M:
            n_prime = self._calc_new_N(y.shape[0])
            new_shape = [n_prime, self.window_size]

            if isinstance(y, np.ndarray):
                res = np.zeros(shape=[n_prime, self.window_size], dtype=object)
                res = res.astype(dtype)
            else:
                import torch
                if isinstance(y, torch.Tensor):
                    res = torch.zeros(new_shape, dtype=torch.int64)
                else:
                    raise ValueError('X has to be either ndarray or of type torch tensor')

            for r, i in enumerate(range(0, n_prime*self.stride, self.stride)):
                res[r, :] = y[i: i+self.window_size]
            return res.squeeze()

        elif self.rep == self.REP_M2O:
            res = y[list(range(self.window_size-1, y.shape[0], self.stride))]
            if isinstance(y, np.ndarray):
                res = res.squeeze().astype(dtype)
                return res
            else:
                import torch
                raise NotImplementedError
                #return res
        else:
            raise ValueError


from pyadlml.constants import ACTIVITY, START_TIME, END_TIME

def _style_colorbar(fig, label):
    fig.update_traces(colorbar=dict(
        title=dict(text=label, font=dict(size=10)),
        titleside='right',
        thickness=10,
    ))

def _dyn_y_label_size(plot_height, nr_labels):

    if nr_labels < 15:
        return 12
    elif nr_labels < 20:
        return 11
    elif nr_labels < 30:
        return 10
    else:
        return 9
import pandas as pd

class ActivityDict(dict):
    """ Dictionary with activity pd.DataFrames as values and subject names as keys.
    """
    def nr_acts(self):
        """"""
        return max([len(df_acts[ACTIVITY].unique()) for df_acts in self.values()])

    def get_activity_union(self): 
        return list(set([item for v in self.values() \
                              for item in v[ACTIVITY].unique()]))

    def get_min_start_time(self):
        return min([df_acts[START_TIME].iloc[0] for df_acts in self.values()])

    def get_max_end_time(self):
        return min([df_acts[END_TIME].iloc[-1] for df_acts in self.values()])

    def concat(self):
        return pd.concat(self.values())


import pandas as pd
from pathlib import Path
from pyadlml.constants import ACTIVITY, DEVICE, START_TIME, END_TIME
from pyadlml.dataset._core.activities import _create_activity_df
from pyadlml.dataset._core.devices import device_boolean_on_states_to_events
from pyadlml.dataset._core.obj import Data
from pyadlml.dataset.io import fetcher, correct_acts_and_devs, download_from_mega

MITLAB_URL = 'https://mega.nz/file/MB4BFL6S#8MjAQoS-j0Lje1UFoWUMOCay2FcdpVfla6p9MTe4SQM'
MITLAB_FILENAME = 'mitlab.zip'

SUB1 = 'subject1'
SUB2 = 'subject2'

@correct_acts_and_devs
@fetcher('mitlab', download_from_mega, MITLAB_FILENAME, MITLAB_URL)
def fetch_mitlab(subject='subject1', keep_original=False, cache=True,
                 retain_corrections=False, folder_path=None):
    """
    Fetches the :ref:`mitlab <ds_mitlab>` dataset from the internet. The original dataset or its cached version
    is stored in the :ref:`data home <storage>` folder.

    Parameters
    ----------
    subject : str of {'subject1', 'subject2'}
        Determines which dataset is loaded.
    keep_original : bool, default=True
        Determines whether the original dataset is deleted after downloading
        or kept on the hard drive.
    cache : bool, default=True
        Determines whether the data object should be stored as a binary file for quicker access.
        For more information how caching is used refer to the :ref:`user guide <storage>`.
    retain_corrections : bool, default=False
        When set to *true*, data points that change or drop during preprocessing
        are listed in respective attributes of the data object. Fore more information
        about error correction refer to the :ref:`user guide <error_correction>`.

    Returns
    -------
    data : object
    """
    assert subject in [SUB1, SUB2]

    act_path = Path(folder_path).joinpath(subject, "Activities.csv")
    dev_path = Path(folder_path).joinpath(subject, "sensors.csv")
    data_path = Path(folder_path).joinpath(subject, "activities_data.csv")

    df_dev_map = _load_device_map(dev_path)
    df_act_map = _load_activity_map(act_path)
    df_dev, df_act = _read_data(data_path, df_dev_map, df_act_map)
    df_dev = device_boolean_on_states_to_events(df_dev)

    lst_act = df_act[ACTIVITY].unique()
    lst_dev = df_dev[DEVICE].unique()

    data = Data(df_act, df_dev, activity_list=lst_act, device_list=lst_dev)

    return data


def _load_device_map(path_to_file):
    df_subx_dev = pd.read_csv(path_to_file, sep=",", header=None)
    df_subx_dev.columns = ['id', 'room', 'device']

    df_subx_dev['device'] = df_subx_dev['id'].astype(str) + ' - ' \
                            + df_subx_dev['room'] + ' ' + df_subx_dev['device']
    df_subx_dev = df_subx_dev.drop(columns='room')
    df_subx_dev = df_subx_dev.set_index('id')
    return df_subx_dev


def _load_activity_map(path_to_file):
    return pd.read_csv(path_to_file, sep=",")


def _read_data(path_to_file, df_dev, df_act):
    """ creates the device dataframe and activity dataframe
    
    The data is present in the following format:
         ACTIVITY_LABEL,DATE,START_TIME,END_TIME
         SENSOR1_ID, SENSOR2_ID, ......
         SENSOR1_OBJECT,SENSOR2_OBJECT, .....
         SENSOR1_ACTIVATION_TIME,SENSOR2_ACTIVATION_TIME, .....
         SENSOR1_DEACTIVATION_TIME,SENSOR2_DEACTIVATION_TIME, .....

         where date is in the mm/dd/yyyy format
         where time is in the hh:mm:ss format
    Parameters
    ----------
    path_to_file: str
    
    Returns
    -------
    df_devices : pd.DataFrame
    
    df_activities : pd.DataFrame
        
    """
    # create empy dataframes for devices and activities
    df_devices = pd.DataFrame(columns=[START_TIME, END_TIME, DEVICE])
    df_activities = _create_activity_df()

    act_list = df_act['Subcategory'].values
    
    with open(path_to_file, 'r') as f_o:
        i = 0
        read_in_device = False
        date = None
        for line in f_o.readlines():
            assert i in [0,1,2,3]
            s = line.split(',')
            # check if the line represents an activity
            if s[0] in act_list:
                assert len(s) == 4
                date = s[1]
                """
                    there is an error where the date is 4/31/2003 which doesn't exist in 
                    subject 2 data. Convert this to the next day 
                """
                if date == '4/31/2003':
                    date = '5/1/2003'
                new_row = {'activity':s[0], 
                           'start_time':pd.Timestamp(date +'T'+s[2]), 
                           'end_time':pd.Timestamp(date +'T'+s[3])
                          }
                df_activities = df_activities.append(new_row, ignore_index=True)
                continue      
                
            # check if the line represents a device
            elif not read_in_device:
                try:
                    read_in_device = True
                    devices = s
                    # delete the '\n' for every last device 
                    devices[-1:] = [devices[-1:][0][:-1]]
                    i = 1
                    continue
                except:
                    raise ValueError
            elif i == 1:
                i = 2
                continue
            elif i == 2:
                ts_act = s                
                i = 3
            elif i == 3:
                ts_deact = s
                
                assert len(ts_act) == len(ts_deact)
                # correct timestamps by inserting a 0 where only a single digit is present
                for j in range(len(ts_act)):
                    if len(ts_act[j]) != 8:
                        ts = ts_act[j].split(':')
                        for k, digit in enumerate(ts):
                            if len(digit) == 1:
                                ts[k] = '0' + ts[k]
                        ts_act[j] = ':'.join(ts)
                    if len(ts_deact) != 8:
                        ts = ts_deact[j].split(':')
                        for k, digit in enumerate(ts):
                            if len(digit) == 1:
                                ts[k] = '0' + ts[k]
                        ts_deact[j] = ':'.join(ts)
                        
                # create rows
                for dev, ts_start, ts_end in zip(devices, ts_act, ts_deact):
                    new_row = {DEVICE:dev,
                               START_TIME:pd.Timestamp(date +'T' + ts_start),
                               END_TIME:pd.Timestamp(date +'T' + ts_end)
                              }
                    df_devices = df_devices.append(new_row, ignore_index=True)                
                i = 0
                read_in_device = False
                
        f_o.close()
        
    # map device ids to strings    
    df_devices[DEVICE] = df_devices[DEVICE].astype(int)
    df_devices[DEVICE] = df_devices[DEVICE].map(df_dev.to_dict()[DEVICE])
    df_devices = df_devices.sort_values(by=START_TIME)
    df_devices = df_devices.drop_duplicates().reset_index(drop=True)
    
    df_activities = df_activities.sort_values(by=START_TIME).reset_index(drop=True)

    return df_devices, df_activities



import numbers
import time
import warnings
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from contextlib import suppress
from functools import partial
from traceback import format_exc
import numpy as np

from joblib import Parallel, delayed, logger
from itertools import product
from sklearn.base import MetaEstimatorMixin, BaseEstimator, is_classifier, clone, _is_pairwise
from sklearn.exceptions import NotFittedError, FitFailedWarning
from sklearn.metrics import check_scoring
from sklearn.metrics._scorer import _check_multimetric_scoring, _MultimetricScorer
from sklearn.model_selection import check_cv
from sklearn.model_selection._search import _check_param_grid, ParameterGrid, _normalize_score_results
from sklearn.model_selection._validation import _aggregate_score_dicts, _score
from sklearn.utils import _deprecate_positional_args, _message_with_time, _safe_indexing
from sklearn.utils.fixes import MaskedArray
from sklearn.utils.metaestimators import if_delegate_has_method# ,_safe_split
from sklearn.utils.validation import check_is_fitted, indexable, _check_fit_params, _num_samples

from pyadlml.pipeline import EvalOnlyWrapper, TrainOnlyWrapper, Pipeline, TrainOrEvalOnlyWrapper
from pyadlml.preprocessing import CVSubset
from sklearn.model_selection._search import BaseSearchCV as SklearnBaseSearchCV

class BaseSearchCV(SklearnBaseSearchCV):
    """Abstract base class for hyper parameter search with cross-validation.
    """

    @abstractmethod
    @_deprecate_positional_args
    def __init__(self, estimator, *, scoring=None, n_jobs=None,
                 online_train_val_split=False,
                 refit=True, cv=None, verbose=0,
                 pre_dispatch='2*n_jobs', error_score=np.nan,
                 return_train_score=True):

        SklearnBaseSearchCV.__init__(self, estimator=estimator, scoring=scoring, n_jobs=n_jobs,
                 refit=refit, cv=cv, verbose=verbose,
                 pre_dispatch=pre_dispatch, error_score=error_score,
                 return_train_score=return_train_score)

        self.online_train_val_split = online_train_val_split

    def score(self, X, y=None):
        """Returns the score on the given data, if the estimator has been refit.
        This uses the score defined by ``scoring`` where provided, and the
        ``best_estimator_.score`` method otherwise.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input data, where n_samples is the number of samples and
            n_features is the number of features.
        y : array-like of shape (n_samples, n_output) \
            or (n_samples,), default=None
            Target relative to X for classification or regression;
            None for unsupervised learning.
        Returns
        -------
        score : float
        """
        self._check_is_fitted('score')
        if self.scorer_ is None:
            raise ValueError("No score function explicitly defined, "
                             "and the estimator doesn't provide one %s"
                             % self.best_estimator_)
        if isinstance(self.scorer_, dict):
            if self.multimetric_:
                scorer = self.scorer_[self.refit]
            else:
                scorer = self.scorer_
            return scorer(self.best_estimator_, X, y)

        # callable
        score = self.scorer_(self.best_estimator_, X, y)
        if self.multimetric_:
            score = score[self.refit]
        return score

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def score_samples(self, X):
        """Call score_samples on the estimator with the best found parameters.
        Only available if ``refit=True`` and the underlying estimator supports
        ``score_samples``.
        .. versionadded:: 0.24
        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements
            of the underlying estimator.
        Returns
        -------
        y_score : ndarray of shape (n_samples,)
        """
        self._check_is_fitted('score_samples')
        return self.best_estimator_.score_samples(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict(self, X):
        """Call predict on the estimator with the best found parameters.
        Only available if ``refit=True`` and the underlying estimator supports
        ``predict``.
        Parameters
        ----------
        X : indexable, length n_samples
            Must fulfill the input assumptions of the
            underlying estimator.
        """
        self._check_is_fitted('predict')
        return self.best_estimator_.predict(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict_proba(self, X):
        """Call predict_proba on the estimator with the best found parameters.
        Only available if ``refit=True`` and the underlying estimator supports
        ``predict_proba``.
        Parameters
        ----------
        X : indexable, length n_samples
            Must fulfill the input assumptions of the
            underlying estimator.
        """
        self._check_is_fitted('predict_proba')
        return self.best_estimator_.predict_proba(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def predict_log_proba(self, X):
        """Call predict_log_proba on the estimator with the best found parameters.
        Only available if ``refit=True`` and the underlying estimator supports
        ``predict_log_proba``.
        Parameters
        ----------
        X : indexable, length n_samples
            Must fulfill the input assumptions of the
            underlying estimator.
        """
        self._check_is_fitted('predict_log_proba')
        return self.best_estimator_.predict_log_proba(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def decision_function(self, X):
        """Call decision_function on the estimator with the best found parameters.
        Only available if ``refit=True`` and the underlying estimator supports
        ``decision_function``.
        Parameters
        ----------
        X : indexable, length n_samples
            Must fulfill the input assumptions of the
            underlying estimator.
        """
        self._check_is_fitted('decision_function')
        return self.best_estimator_.decision_function(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def transform(self, X):
        """Call transform on the estimator with the best found parameters.
        Only available if the underlying estimator supports ``transform`` and
        ``refit=True``.
        Parameters
        ----------
        X : indexable, length n_samples
            Must fulfill the input assumptions of the
            underlying estimator.
        """
        self._check_is_fitted('transform')
        return self.best_estimator_.transform(X)

    @if_delegate_has_method(delegate=('best_estimator_', 'estimator'))
    def inverse_transform(self, Xt):
        """Call inverse_transform on the estimator with the best found params.
        Only available if the underlying estimator implements
        ``inverse_transform`` and ``refit=True``.
        Parameters
        ----------
        Xt : indexable, length n_samples
            Must fulfill the input assumptions of the
            underlying estimator.
        """
        self._check_is_fitted('inverse_transform')
        return self.best_estimator_.inverse_transform(Xt)

    def fit(self, X, y=None, *, groups=None, **fit_params):
        """Run fit with all sets of parameters.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples and
            n_features is the number of features.
        y : array-like of shape (n_samples, n_output) \
            or (n_samples,), default=None
            Target relative to X for classification or regression;
            None for unsupervised learning.
        groups : array-like of shape (n_samples,), default=None
            Group labels for the samples used while splitting the dataset into
            train/test set. Only used in conjunction with a "Group" :term:`cv`
            instance (e.g., :class:`~sklearn.model_selection.GroupKFold`).
        **fit_params : dict of str -> object
            Parameters passed to the ``fit`` method of the estimator
        """
        estimator = self.estimator
        refit_metric = "score"

        if callable(self.scoring):
            scorers = self.scoring
        elif self.scoring is None or isinstance(self.scoring, str):
            scorers = check_scoring(self.estimator, self.scoring)
        else:
            scorers = _check_multimetric_scoring(self.estimator, self.scoring)
            self._check_refit_for_multimetric(scorers)
            refit_metric = self.refit

        #X, y, groups = indexable(X, y, groups) # todo debug
        fit_params = _check_fit_params(X, fit_params)

        cv_orig = check_cv(self.cv, y, classifier=is_classifier(estimator))
        n_splits = cv_orig.get_n_splits(X, y, groups)

        base_estimator = clone(self.estimator)

        parallel = Parallel(n_jobs=self.n_jobs,
                            pre_dispatch=self.pre_dispatch)

        fit_and_score_kwargs = dict(scorer=scorers,
                                    fit_params=fit_params,
                                    return_train_score=self.return_train_score,
                                    return_n_test_samples=True,
                                    return_times=True,
                                    return_parameters=False,
                                    error_score=self.error_score,
                                    verbose=self.verbose)
        results = {}
        with parallel:
            all_candidate_params = []
            all_out = []
            all_more_results = defaultdict(list)

            def evaluate_candidates(candidate_params, cv=None,
                                    more_results=None):
                cv = cv or cv_orig
                candidate_params = list(candidate_params)
                n_candidates = len(candidate_params)

                if self.verbose > 0:
                    print("Fitting {0} folds for each of {1} candidates,"
                          " totalling {2} fits".format(
                              n_splits, n_candidates, n_candidates * n_splits))

                if self.online_train_val_split:
                    can = enumerate(candidate_params)
                    spl = enumerate(cv.split(X, None, groups))
                    lst = []
                    for (cand_idx, parameters), (split_idx, (train, test)) in product(can, spl):
                        lst.append(delayed(_fit_and_score)(
                            clone(base_estimator),
                            X, y,
                            train=train, test=test,
                            parameters=parameters,
                            online_train_val_split=True,
                            **fit_and_score_kwargs))
                    out = parallel(lst)
                else:
                    can = enumerate(candidate_params)
                    spl = enumerate(cv.split(X, y, groups))
                    lst = []
                    for (cand_idx, parameters), (split_idx, (train, test)) in product(can, spl):
                        lst.append(delayed(_fit_and_score)(
                            clone(base_estimator),
                            X, y,
                            train=train, test=test,
                            parameters=parameters,
                            split_progress=(
                               split_idx,
                               n_splits),
                            candidate_progress=(
                               cand_idx,
                               n_candidates),
                            online_train_val_split=False,
                            **fit_and_score_kwargs))
                    out = parallel(lst)

                if len(out) < 1:
                    raise ValueError('No fits were performed. '
                                     'Was the CV iterator empty? '
                                     'Were there no candidates?')
                elif len(out) != n_candidates * n_splits:
                    raise ValueError('cv.split and cv.get_n_splits returned '
                                     'inconsistent results. Expected {} '
                                     'splits, got {}'
                                     .format(n_splits,
                                             len(out) // n_candidates))

                # For callable self.scoring, the return type is only know after
                # calling. If the return type is a dictionary, the error scores
                # can now be inserted with the correct key. The type checking
                # of out will be done in `_insert_error_scores`.
                if callable(self.scoring):
                    _insert_error_scores(out, self.error_score)
                all_candidate_params.extend(candidate_params)
                all_out.extend(out)
                if more_results is not None:
                    for key, value in more_results.items():
                        all_more_results[key].extend(value)

                nonlocal results
                results = self._format_results(
                    all_candidate_params, n_splits, all_out,
                    all_more_results)

                return results

            self._run_search(evaluate_candidates)

            # multimetric is determined here because in the case of a callable
            # self.scoring the return type is only known after calling
            first_test_score = all_out[0]['test_scores']
            self.multimetric_ = isinstance(first_test_score, dict)

            # check refit_metric now for a callabe scorer that is multimetric
            if callable(self.scoring) and self.multimetric_:
                self._check_refit_for_multimetric(first_test_score)
                refit_metric = self.refit

        # For multi-metric evaluation, store the best_index_, best_params_ and
        # best_score_ iff refit is one of the scorer names
        # In single metric evaluation, refit_metric is "score"
        if self.refit or not self.multimetric_:
            # If callable, refit is expected to return the index of the best
            # parameter set.
            if callable(self.refit):
                self.best_index_ = self.refit(results)
                if not isinstance(self.best_index_, np.numbers.Integral):
                    raise TypeError('best_index_ returned is not an integer')
                if (self.best_index_ < 0 or
                   self.best_index_ >= len(results["params"])):
                    raise IndexError('best_index_ index out of range')
            else:
                self.best_index_ = results["rank_test_%s"
                                           % refit_metric].argmin()
                self.best_score_ = results["mean_test_%s" % refit_metric][
                                           self.best_index_]
            self.best_params_ = results["params"][self.best_index_]

        if self.refit:
            # we clone again after setting params in case some
            # of the params are estimators as well.
            self.best_estimator_ = clone(clone(base_estimator).set_params(
                **self.best_params_))

            refit_start_time = time.time()
            if isinstance(self.best_estimator_, Pipeline):
                self.best_estimator_.train()
                # set the cross val splitter training range to the whole dataset
                if self.online_train_val_split:
                    for estim in self.best_estimator_:
                        if isinstance(estim, CVSubset) and isinstance(estim, TrainOnlyWrapper):
                            tmp = np.arange(len(X))
                            estim.set_range(tmp)
                            break

            if y is not None:
                self.best_estimator_.fit(X, y, **fit_params)
            else:
                self.best_estimator_.fit(X, **fit_params)

            if isinstance(self.best_estimator_, Pipeline):
                self.best_estimator_.prod()
            refit_end_time = time.time()
            self.refit_time_ = refit_end_time - refit_start_time

        # Store the only scorer not as a dict for single metric evaluation
        self.scorer_ = scorers

        self.cv_results_ = results
        self.n_splits_ = n_splits

        return self

class GridSearchCV(BaseSearchCV):
    """Exhaustive search over specified parameter values for an estimator.
    """
    _required_parameters = ["estimator", "param_grid"]

    @_deprecate_positional_args
    def __init__(self, estimator, param_grid, *, online_train_val_split=False,
                 scoring=None, n_jobs=None, refit=True, cv=None,
                 verbose=0, pre_dispatch='2*n_jobs',
                 error_score=np.nan, return_train_score=False):
        super().__init__(
            estimator=estimator, scoring=scoring,
            online_train_val_split=online_train_val_split,
            n_jobs=n_jobs, refit=refit, cv=cv, verbose=verbose,
            pre_dispatch=pre_dispatch, error_score=error_score,
            return_train_score=return_train_score)
        self.param_grid = param_grid
        _check_param_grid(param_grid)

    def _run_search(self, evaluate_candidates):
        """Search all candidates in param_grid"""
        evaluate_candidates(ParameterGrid(self.param_grid))



def _fit_and_score(estimator, X, y, scorer, train, test, verbose,
                   parameters, fit_params, return_train_score=False,
                   return_parameters=False, return_n_test_samples=False,
                   return_times=False, return_estimator=False,
                   split_progress=None, candidate_progress=None,
                   error_score=np.nan, online_train_val_split=False):

    """Fit estimator and compute scores for a given dataset split.
    Parameters
    ----------
    estimator : estimator object implementing 'fit'
        The object to use to fit the data.
    X : array-like of shape (n_samples, n_features)
        The data to fit.
    y : array-like of shape (n_samples,) or (n_samples, n_outputs) or None
        The target variable to try to predict in the case of
        supervised learning.
    scorer : A single callable or dict mapping scorer name to the callable
        If it is a single callable, the return value for ``train_scores`` and
        ``test_scores`` is a single float.
        For a dict, it should be one mapping the scorer name to the scorer
        callable object / function.
        The callable object / fn should have signature
        ``scorer(estimator, X, y)``.
    train : array-like of shape (n_train_samples,)
        Indices of training samples.
    test : array-like of shape (n_test_samples,)
        Indices of test samples.
    verbose : int
        The verbosity level.
    error_score : 'raise' or numeric, default=np.nan
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised.
        If a numeric value is given, FitFailedWarning is raised.
    parameters : dict or None
        Parameters to be set on the estimator.
    fit_params : dict or None
        Parameters that will be passed to ``estimator.fit``.
    return_train_score : bool, default=False
        Compute and return score on training set.
    return_parameters : bool, default=False
        Return parameters that has been used for the estimator.
    split_progress : {list, tuple} of int, default=None
        A list or tuple of format (<current_split_id>, <total_num_of_splits>).
    candidate_progress : {list, tuple} of int, default=None
        A list or tuple of format
        (<current_candidate_id>, <total_number_of_candidates>).
    return_n_test_samples : bool, default=False
        Whether to return the ``n_test_samples``.
    return_times : bool, default=False
        Whether to return the fit/score times.
    return_estimator : bool, default=False
        Whether to return the fitted estimator.
    Returns
    -------
    result : dict with the following attributes
        train_scores : dict of scorer name -> float
            Score on training set (for all the scorers),
            returned only if `return_train_score` is `True`.
        test_scores : dict of scorer name -> float
            Score on testing set (for all the scorers).
        n_test_samples : int
            Number of test samples.
        fit_time : float
            Time spent for fitting in seconds.
        score_time : float
            Time spent for scoring in seconds.
        parameters : dict or None
            The parameters that have been evaluated.
        estimator : estimator object
            The fitted estimator.
        fit_failed : bool
            The estimator failed to fit.
    """
    if not isinstance(error_score, numbers.Number) and error_score != 'raise':
        raise ValueError(
            "error_score must be the string 'raise' or a numeric value. "
            "(Hint: if using 'raise', please make sure that it has been "
            "spelled correctly.)"
        )

    progress_msg = ""
    if verbose > 2:
        if split_progress is not None:
            progress_msg = f" {split_progress[0]+1}/{split_progress[1]}"
        if candidate_progress and verbose > 9:
            progress_msg += (f"; {candidate_progress[0]+1}/"
                             f"{candidate_progress[1]}")

    if verbose > 1:
        if parameters is None:
            params_msg = ''
        else:
            sorted_keys = sorted(parameters)  # Ensure deterministic o/p
            params_msg = (', '.join(f'{k}={parameters[k]}'
                                    for k in sorted_keys))
    if verbose > 9:
        start_msg = f"[CV{progress_msg}] START {params_msg}"
        print(f"{start_msg}{(80 - len(start_msg)) * '.'}")

    # Adjust length of sample weights
    fit_params = fit_params if fit_params is not None else {}
    fit_params = _check_fit_params(X, fit_params, train)

    if parameters is not None:
        # clone after setting parameters in case any parameters
        # are estimators (like pipeline steps)
        # because pipeline doesn't clone steps in fit
        cloned_parameters = {}
        for k, v in parameters.items():
            cloned_parameters[k] = clone(v, safe=False)

        estimator = estimator.set_params(**cloned_parameters)

    start_time = time.time()


    if online_train_val_split:
        # inject the train and test data into the corresponding Subset selectors
        set_train_estim = False
        set_test_estim = False
        for estim in estimator:
            if set_train_estim and set_test_estim:
                break
            if isinstance(estim, CVSubset) and isinstance(estim, EvalOnlyWrapper):
                estim.set_range(test)
                set_test_estim = True
            if isinstance(estim, CVSubset) and isinstance(estim, TrainOnlyWrapper):
                estim.set_range(train)
                set_train_estim = True
            # TODO delete unti 906 above
            if isinstance(estim, CVOnlineSplitter) and isinstance(estim, TrainOrEvalOnlyWrapper):
                estim.set_train_range(train)
                estim.set_test_range(test)
                set_train_estim = True
                set_test_estim = True
                break
        if not set_train_estim or not set_test_estim:
            raise ValueError("when specifying online learning a KeepTrain and KeepTest have to be in the pipeline")
    else:
        X_train, y_train = _safe_split(estimator, X, y, train)
        X_test, y_test = _safe_split(estimator, X, y, test, train)

    result = {}
    try:
        if online_train_val_split:
            estimator = estimator.train()
            estimator.fit(X, y, **fit_params)
        else:
            if y_train is None:
                estimator.fit(X_train, **fit_params)
            else:
                estimator.fit(X_train, y_train, **fit_params)

    except Exception as e:
        # Note fit time as time until error
        fit_time = time.time() - start_time
        score_time = 0.0
        if error_score == 'raise':
            raise
        elif isinstance(error_score, numbers.Number):
            if isinstance(scorer, dict):
                test_scores = {name: error_score for name in scorer}
                if return_train_score:
                    train_scores = test_scores.copy()
            else:
                test_scores = error_score
                if return_train_score:
                    train_scores = error_score
            warnings.warn("Estimator fit failed. The score on this train-test"
                          " partition for these parameters will be set to %f. "
                          "Details: \n%s" %
                          (error_score, format_exc()),
                          FitFailedWarning)
        result["fit_failed"] = True
        y_sample_len = len(test)
    else:
        result["fit_failed"] = False

        fit_time = time.time() - start_time


        estimator.eval()
        if online_train_val_split:
            # select estimator without the classifier and transform x and y
            # to retrieve y_test
            try:
                X_prime, y_prime = estimator[:-1].transform(X, y)
            except Exception:
                print('sth. went terribly wrong')
                raise
            y_sample_len = len(y_prime)
            test_scores = _score(estimator[-1], X_prime, y_prime, scorer)
            # line above also catches the case when samples are dropped for x and y after
            # the validation set selection for e.g DropNanTransformer()
        else:
            test_scores = _score(estimator, X_test, y_test, scorer, error_score)
        score_time = time.time() - start_time - fit_time
        if return_train_score:
            if online_train_val_split:
                estimator.train()

                _, y_prime = estimator[:-1].transform(X, y)
                if isinstance(y_prime, pd.DataFrame) and len(y_prime.columns) == 1:
                    y_prime = y_prime.T.values.squeeze()
                train_scores = _score(estimator, X, y_prime, scorer)

                estimator.eval()
            else:
                train_scores = _score(
                    estimator, X_train, y_train, scorer, error_score
                )

    if verbose > 1:
        total_time = score_time + fit_time
        end_msg = f"[CV{progress_msg}] END "
        result_msg = params_msg + (";" if params_msg else "")
        if verbose > 2 and isinstance(test_scores, dict):
            for scorer_name in sorted(test_scores):
                result_msg += f" {scorer_name}: ("
                if return_train_score:
                    scorer_scores = train_scores[scorer_name]
                    result_msg += f"train={scorer_scores:.3f}, "
                result_msg += f"test={test_scores[scorer_name]:.3f})"
        result_msg += f" total time={logger.short_format_time(total_time)}"

        # Right align the result_msg
        end_msg += "." * (80 - len(end_msg) - len(result_msg))
        end_msg += result_msg
        print(end_msg)

    result["test_scores"] = test_scores
    if return_train_score:
        result["train_scores"] = train_scores
    if return_n_test_samples:
        if online_train_val_split:
            result["n_test_samples"] = y_sample_len
        else:
            result["n_test_samples"] = _num_samples(X_test)
    if return_times:
        result["fit_time"] = fit_time
        result["score_time"] = score_time
    if return_parameters:
        result["parameters"] = parameters
    if return_estimator:
        result["estimator"] = estimator
    return result


def _insert_error_scores(results, error_score):
    """Insert error in `results` by replacing them inplace with `error_score`.
    This only applies to multimetric scores because `_fit_and_score` will
    handle the single metric case.
    """
    successful_score = None
    failed_indices = []
    for i, result in enumerate(results):
        if result["fit_failed"]:
            failed_indices.append(i)
        elif successful_score is None:
            successful_score = result["test_scores"]

    if successful_score is None:
        raise NotFittedError("All estimators failed to fit")

    if isinstance(successful_score, dict):
        formatted_error = {name: error_score for name in successful_score}
        for i in failed_indices:
            results[i]["test_scores"] = formatted_error.copy()
            if "train_scores" in results[i]:
                results[i]["train_scores"] = formatted_error.copy()


def _safe_split(estimator, X, y, indices, train_indices=None):
    """Create subset of dataset and properly handle kernels.

    Slice X, y according to indices for cross-validation, but take care of
    precomputed kernel-matrices or pairwise affinities / distances.

    If ``estimator._pairwise is True``, X needs to be square and
    we slice rows and columns. If ``train_indices`` is not None,
    we slice rows using ``indices`` (assumed the test set) and columns
    using ``train_indices``, indicating the training set.

    .. deprecated:: 0.24

        The _pairwise attribute is deprecated in 0.24. From 1.1
        (renaming of 0.26) and onward, this function will check for the
        pairwise estimator tag.

    Labels y will always be indexed only along the first axis.

    Parameters
    ----------
    estimator : object
        Estimator to determine whether we should slice only rows or rows and
        columns.

    X : array-like, sparse matrix or iterable
        Data to be indexed. If ``estimator._pairwise is True``,
        this needs to be a square array-like or sparse matrix.

    y : array-like, sparse matrix or iterable
        Targets to be indexed.

    indices : array of int
        Rows to select from X and y.
        If ``estimator._pairwise is True`` and ``train_indices is None``
        then ``indices`` will also be used to slice columns.

    train_indices : array of int or None, default=None
        If ``estimator._pairwise is True`` and ``train_indices is not None``,
        then ``train_indices`` will be use to slice the columns of X.

    Returns
    -------
    X_subset : array-like, sparse matrix or list
        Indexed data.

    y_subset : array-like, sparse matrix or list
        Indexed targets.

    """
    if _is_pairwise(estimator):
        if not hasattr(X, "shape"):
            raise ValueError("Precomputed kernels or affinity matrices have "
                             "to be passed as arrays or sparse matrices.")
        # X is a precomputed square kernel matrix
        if X.shape[0] != X.shape[1]:
            raise ValueError("X should be a square kernel matrix")
        if train_indices is None:
            X_subset = X[np.ix_(indices, indices)]
        else:
            X_subset = X[np.ix_(indices, train_indices)]
    else:
        X_subset = _safe_indexing(X, indices)

    if y is not None:
        y_subset = _safe_indexing(y, indices)
    else:
        y_subset = None

    return X_subset, y_subset


from pyadlml.constants import DEVICE, ACTIVITY
from pyadlml.dataset.io import download_from_mega, get_data_home, _ensure_dh_folder_exists, _delete_data
from pathlib import Path
import pandas as pd
import joblib


class Data(object):
    def __init__(self, activities=None, devices=None, activity_list=None, device_list=None):

        self.df_devices = devices

        if device_list is None:
            self.lst_devices = devices[DEVICE].unique()

        if activities is not None:
            self.df_activities = activities
            if activity_list is None:
                self.lst_activities = activities[ACTIVITY].unique()
            else:
                self.lst_activities = activity_list
            self.act_dict = {"default": self.df_activities}
        else:
            self.act_dict = {}

    def set_activity_list(self, lst, name=None):
        """
        """
        if name is None or name == "default":
            setattr(self, "lst_activities", lst)
        else:
            setattr(self, "lst_activities_{}".format(name), lst)

    def set_activity_df(self, df, name=None):
        """
        """
        if name is None or name == "default":
            setattr(self, "df_activities", df)
            self.act_dict["default"] = df
        else:
            setattr(self, "df_activities_{}".format(name), df)
            self.act_dict[name] = df

    def set_activity_correction(self, lst, name=None):
        if name is None or name == "default":
            self.correction_activities = lst
        else:
            setattr(self, "correction_activities_{}".format(name), lst)

    def set_device_correction(self, corr_dts, corr_incons):
        self.correction_devices_duplicate_timestamps = corr_dts
        self.correction_devices_on_off_inconsistency = corr_incons

    def get_activity_dict(self) -> dict:
        return self.act_dict

    def set_device_df(self, df):
        self.df_devices = df

from abc import ABC, abstractmethod

class DatasetDownloader():
    def __init__(self,):
        pass



class DataFetcher(ABC):
    FN_CLEANED = 'cleaned.joblib'
    FN_CACHED = 'cached.joblib'

    def __init__(self, dataset_name: str,
                       downloader: DatasetDownloader,
                       correct_activities=False,
                       correct_devices=False
        ):

        self.downloader = downloader
        self.ds_name = dataset_name
        self.correct_devices = correct_devices
        self.correct_activities = correct_activities


        self.data_home = get_data_home()
        self._dataset_folder = None
        self._original_folder = None


    def _create_exp_folder(self):
        """
        """
        _ensure_dh_folder_exists()
        self.data_home = get_data_home()

    @property
    def dataset_folder(self):
        return Path(self.data_home).joinpath(self.ds_name)

    @property
    def original_folder(self):
        return self.dataset_folder.joinpath('original')



    def apply_corrections(self, data:dict, retain_corrections=False):
        """ Applies corrections

        data : dict
            Of the form 
            {
                'activity_list': [],
                'device_list': [],
                'df_devices': pd.DataFrame,
                'df_activities_subject_M': pd.DataFrame,
                ...
            }
        """
        for key, df in data.items():
            if 'df_activity' in key: 
                from pyadlml.dataset._core.activities import correct_activities
                if self.correct_activities:
                    df, corr = correct_activities(df, retain_corrections)
                    data[key] = df
                    if retain_corrections:
                        data[key + "_corrections"] = corr
                else:
                    data[key] = df

        if self.correct_devices:
            from pyadlml.dataset._core.devices import correct_devices
            df_dev, correction_dev_dict = correct_devices(data['df_devices'], retain_correction=retain_corrections)
            data['df_devices'] = df_dev
            if retain_corrections:
                data[key + "_corrections"] = corr

        return data


    def __call__(self, cache=False, keep_original=False, retain_corrections=False, load_cleaned=False, folder_path=None, *args, **kwargs):

        # Resolve folders
        self._create_exp_folder()

        sel = None
        for a in kwargs.keys():
            sel = a

        if sel is not None:
            # TODO, when do I get here
            cached_name = FN_CACHED.split(
                '.')[0] + '_' + kwargs[sel] + '.' + FN_CACHED.split('.')[1]
            cleaned_name = FN_CLEANED.split(
                '.')[0] + '_' + kwargs[sel] + '.' + FN_CLEANED.split('.')[1]
            fp_cached_dataset = self.dataset_folder.joinpath(cached_name)
            fp_cleaned_dataset = self.dataset_folder.joinpath(cleaned_name)
        else:
            fp_cached_dataset = self.dataset_folder.joinpath(self.FN_CACHED)
            fp_cleaned_dataset = self.dataset_folder.joinpath(self.FN_CLEANED)


        # Only download data if it was not already fetched
        # and no cached version exists that should be loaded
        # and the load_cleaned flag is not set
        if not (self.original_folder.exists() or (fp_cached_dataset.is_file() and cache)
                or folder_path is not None) and not load_cleaned:
            self.original_folder.mkdir(parents=True, exist_ok=True)
            self.download()

        elif load_cleaned and not fp_cleaned_dataset.exists():
            self.download_cleaned(self.dataset_folder)


        # Load from cached version if available and cache flag is set
            # or load from cleaned version
            # otherwise load from the function
        if fp_cached_dataset.is_file() and cache and not load_cleaned:
            data = joblib.load(fp_cached_dataset)
        elif fp_cleaned_dataset.is_file() and load_cleaned and not cache:
            data = joblib.load(fp_cleaned_dataset)
        else:
            data = self.load_data(folder_path=self.original_folder, **kwargs)

        # Save dataset if downloaded and not already cached
        if cache and not fp_cached_dataset.is_file() and not load_cleaned:
            joblib.dump(data, fp_cached_dataset)

        # Clean up data
        if not cache and fp_cached_dataset.is_file():
            fp_cached_dataset.unlink()
        if not keep_original and self.original_folder.exists():
            _delete_data(self.original_folder)

        # Correct devices or activities
        self.apply_corrections(data)

        return data

    def download(self):
        self.downloader.download(self.original_folder)

    def download_cleaned(self):
        self.downloader.download_cleaned(self.dataset_folder)


    @abstractmethod
    def load_data(self, *args, **kwargs):
        raise NotImplementedError



class MegaDownloader(DatasetDownloader):
    def __init__(self, url, fn, url_cleaned, fn_cleaned):
        self.mega_url =  url
        self.fn = fn
        self.url_cleaned = url_cleaned
        self.fn_cleaned = fn_cleaned
    
    
    def download_cleaned(self, dest: Path) -> None:
        """ Download from mega"""
        download_from_mega(dest, self.fn_cleaned, self.url_cleaned, unzip=False)

    def download(self, dest: Path) -> None:
        download_from_mega(dest, self.fn, self.mega_url, unzip=True)



from skorch.utils import to_tensor
from skorch.utils import TeeGenerator
from skorch import NeuralNetClassifier

import torch.nn as nn
import torch.nn.functional as F
import torch

import torch.nn as nn
import torch.nn.functional as F
class RNN(nn.Module):
    def __init__(self, input_size, n_classes, rec_layer_type='lstm',
                 hidden_size=300, hidden_layers=1, seq='many-to-one'):
        super().__init__()
        # set parameters for
        self.input_size = input_size
        self.n_classes = n_classes
        self.hidden_size = hidden_size
        self.hidden_layers = hidden_layers
        self.rec_layer_type = rec_layer_type.lower()
        self.seq_type = seq

        self.reset_weights()

    def reset_weights(self):
        # initialize the weights
        rec_layer = {'lstm': nn.LSTM, 'gru': nn.GRU}[self.rec_layer_type]
        #input and ouput are providedd as (batch, seq, feature)
        self.rnn_ = rec_layer(self.input_size,
                            hidden_size=self.hidden_size,
                            batch_first=True)
        self.out_ = nn.Linear(self.hidden_size, self.n_classes)

    def forward(self, x):
        seq_length = x.shape[1]

        # forward pass through recurrent net
        if self.rec_layer_type == 'gru':
            out, _ = self.rnn_(x)
        elif self.rec_layer_type == 'lstm':
            # output (batch_size , seq length, hidden_size)
            # hn (n_layers, batch_size, hidden_size)
            # cn (n_layers, batch_size, hidden_size)
            out, _ = self.rnn_(x)

        # get either the whole latent state sequence or only the last
        if self.seq_type == 'many-to-one':
            # take output of last stacked RNN layer
            # Note: equiv to latent_rep = h_n[-1]
            latent_rep = out[:, seq_length-1, :]
        else:
            latent_rep = out
            # todo flatten the array
        #print('lr: ', latent_rep.shape)
        out = F.softmax(self.out_(latent_rep), dim=-1)
        #print('X:',out.shape)

        return out

#import skorch
#class RegularizedNet(NeuralNet):
#    def __init__(self, *args, lambda1=0.01, **kwargs):
#        super().__init__(*args, **kwargs)
#        self.lambda1 = lambda1
#
#    def get_loss(self, y_pred, y_true, X=None, training=False):
#        loss = super().get_loss(y_pred, y_true, X=X, training=training)
#        loss += self.lambda1 * sum([w.abs().sum() for w in self.module_.parameters()])
#        return loss


from skorch.utils import to_tensor

class RNNClassifier(NeuralNetClassifier):

    def get_loss(self, y_pred, y_true, X=None, training=False):
        """Return the loss for this batch.
        Parameters
        ----------
        y_pred : torch tensor
          Predicted target values
        y_true : torch tensor
          True target values.
        X : input data, compatible with skorch.dataset.Dataset
        """
        if self.module_.seq_type == 'many-to-many':
            y_true = torch.flatten(to_tensor(y_true, device=self.device), start_dim=0, end_dim=1)
            y_pred = torch.flatten(y_pred, start_dim=0, end_dim=1)

        elif self.module_.seq_type != 'many-to-one':
            ValueError("the sequence type of the RNN was false defined")

        if isinstance(self.criterion_, torch.nn.Module):
            self.criterion_.train(training)

        return self.criterion_(y_pred, y_true)


import pandas as pd
from pyadlml.dataset._core.obj import Data
from pyadlml.constants import ACTIVITY, VALUE, START_TIME, END_TIME, TIME, NAME, DEVICE
import os
from pyadlml.dataset.io import fetcher, correct_acts_and_devs, download_from_mega

CASAS_ARUBA_URL = 'https://mega.nz/file/QA5hEToD#V0ypxFsxiwWgVV49OzhsX8RnMNTX8MYSUM2TLL1xX6w'
CASAS_ARUBA_FILENAME = 'casas_aruba.zip'


@correct_acts_and_devs
@fetcher('casas_aruba', download_from_mega, CASAS_ARUBA_FILENAME, CASAS_ARUBA_URL)
def fetch_casas_aruba(keep_original=True, cache=True, retain_corrections=False,
                      folder_path=None):
    """
    Fetches the casas aruba dataset from the internet. The original dataset or its cached version
    is stored in the :ref:`data home <storage>` folder.

    Parameters
    ----------
    keep_original : bool, default=True
        Determines whether the original dataset is deleted after downloading
        or kept on the hard drive.
    cache : bool, default=True
        Determines whether the data object should be stored as a binary file for quicker access.
        For more information how caching is used refer to the :ref:`user guide <storage>`.
    retain_corrections : bool, default=False
        When set to *true* data points that are changed or dropped during preprocessing
        are listed in the respective attributes of the data object.  Fore more information
        about the attributes refer to the :ref:`user guide <error_correction>`.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_casas_aruba
    >>> data = fetch_casas_aruba()
    >>> dir(data)
    >>> [..., df_activities, df_devices, ...]

    Returns
    -------
    data : object
    """
    _fix_data(os.path.join(folder_path, "data"))
    data_path = os.path.join(folder_path, 'corrected_data.csv')

    df = _load_df(data_path)\
        .drop_duplicates()
    df_dev = _get_devices_df(df)
    df_act = _get_activity_df(df)

    lst_act = df_act[ACTIVITY].unique()
    lst_dev = df_dev[DEVICE].unique()

    return Data(df_act, df_dev, activity_list=lst_act, device_list=lst_dev)


def _fix_data(path):
    """
    as the data is very inconsistent with tabs and spaces this is to make it alright again
    produces: 
        date time,id,value,activity 
    """
    data_path_tf = path[:-4] + 'corrected_data.csv'
    with open(path, 'r') as f_o, open(data_path_tf, 'w') as f_t:
        i= 1
        for line in f_o.readlines():
            s = line.split()
            # weird enterhome and leave home devices that appear inconsistently and are omitted
            if s[2] in ['ENTERHOME', 'LEAVEHOME']:
                continue
            # there is an error in line 1476694 where M014 is replaced with a 'c'
            if s[2] == 'c':
                s[2] = 'M014'
            # 'c' and '5' are randomly added onto values - remove them
            if 'c' in s[3]:
                s[3] = s[3].replace('c', '')                
            if '5' in s[3] and s[2][0] == 'M':
                s[3] = s[3].replace('5', '')
            if s[3] in ['ONM026', 'ONM009', 'ONM024']:         
                s[3] = 'ON'                
            # line 886912 error should be on
            if s[2][0] == 'M' and len(s[3]) == 1:
                s[3] = 'ON'                
            # line 900915 error should be off 
            if s[2][0] == 'M' and len(s[3]) == 2 and s[1] == '18:13:47.291404':
                s[3] = 'OFF'
            # add new line
            new_line = " ".join(s[:2]) + "," + ",".join(s[2:4])
            try:
                s[4] # test if there is an activity
                new_line += "," + " ".join(s[4:])
            except:
                pass
                
            f_t.write(new_line + "\n")
        f_t.close()
        f_o.close()


def _load_df(data_path):
    df = pd.read_csv(data_path,
                     sep=",",
                     #parse_dates=True,
                     infer_datetime_format=True,
                     na_values=True,
                     names=[START_TIME, 'id', VALUE, ACTIVITY],
                     engine='python'  #to ignore warning for fallback to python engine because skipfooter
                     #dtyp
                     )
    df[START_TIME] = pd.to_datetime(df[START_TIME])
    return df

def _val_activity_count(df_act):
    # confirm data assumptions from readme
    assert len(df_act[df_act[ACTIVITY] == 'Meal_Preparation']) == 1606
    # observed 2919 times line below
    len(df_act[df_act[ACTIVITY] == 'Relax'])# == 2910 # does not correspond to authors reported values
    assert len(df_act[df_act[ACTIVITY] == 'Eating']) == 257
    assert len(df_act[df_act[ACTIVITY] == 'Work'])== 171
    assert len(df_act[df_act[ACTIVITY] == 'Sleeping']) == 401
    assert len(df_act[df_act[ACTIVITY] == 'Wash_Dishes']) == 65
    assert len(df_act[df_act[ACTIVITY] == 'Bed_to_Toilet']) == 157
    assert len(df_act[df_act[ACTIVITY] == 'Enter_Home']) == 431
    assert len(df_act[df_act[ACTIVITY] == 'Leave_Home']) == 431
    assert len(df_act[df_act[ACTIVITY] == 'Housekeeping']) == 33
    assert len(df_act[df_act[ACTIVITY] == 'Respirate']) == 6

def _get_devices_df(df):
    df = df.drop(ACTIVITY, axis=1)
    bin_mask = (df[VALUE] == 'ON') | (df[VALUE] == 'OFF')

    # preprocess only binary devices to ON-OFF--> False True
    df_binary = df[bin_mask]
    df_binary[VALUE] = (df_binary[VALUE] == 'ON')

    # preprocess only numeric devices
    num_mask = pd.to_numeric(df[VALUE], errors='coerce').notnull()
    df_num = df[num_mask]
    df_num[VALUE] = df_num[VALUE].astype(float)

    # preprocess categorical devices
    df_cat = df[~num_mask & ~bin_mask]

    # join datasets
    df = pd.concat([df_cat, df_binary, df_num], axis=0, ignore_index=True)
    df.columns = [TIME, DEVICE, VALUE]
    df = df.sort_values(by=TIME).reset_index(drop=True)

    return df


def _get_activity_df(df):
    # get all rows containing activities
    df = df[~df[ACTIVITY].isnull()][[START_TIME, ACTIVITY]].copy()
    
    act_list = pd.unique(df[ACTIVITY])
    
    new_df_lst = []
    for i in range(1, len(act_list), 2):
        activity = act_list[i][:-4]
        act_begin = act_list[i-1]
        act_end = act_list[i]
           
        # create subsets for begin and end of chosen activity
        df_res = df[df[ACTIVITY] == act_begin].reset_index(drop=True)
        df_end = df[df[ACTIVITY] == act_end].reset_index(drop=True)
        
        # append sorted end_time to start_time as they should be
        # pairwise together
        df_res[ACTIVITY] = activity
        df_res[END_TIME] = df_end[START_TIME]
        new_df_lst.append(df_res)
    
    # data preparation
    res = pd.concat(new_df_lst)
    res = res.reindex(columns=[START_TIME, END_TIME, ACTIVITY])
    res = res.sort_values(START_TIME)
    res = res.reset_index(drop=True)
    return res


import functools
import os
from collections import OrderedDict
from pathlib import Path
import inspect

import joblib
import hashlib
import zipfile
import shutil

import pandas as pd

from pyadlml.constants import ACTIVITY, DEVICE
from pyadlml.dataset._core.activities import correct_activities
from pyadlml.dataset._core.devices import correct_devices, CORRECTION_TS, CORRECTION_ONOFF_INCONS
from pyadlml.dataset.util import extract_kwargs

DATA_DUMP_NAME = 'data.joblib'
DATA_HOME_FOLDER_NAME = 'pyadlml_data_home'


def set_data_home(path_to_folder):
    """
    Sets the global variable ``data_home`` and creates the according folder.
    All dump, load and fetch operations assume this folder as the base of operation.

    Parameters
    ----------
    path_to_folder : str
        Specifies the path where the data_home is located.
    """
    from pyadlml import DATA_HOME
    global DATA_HOME
    DATA_HOME[0] = path_to_folder


def _ensure_dh_folder_exists():
    path_to_folder = get_data_home()
    if not os.path.exists(path_to_folder):
        _create_folder(path_to_folder)


def get_data_home():
    """ Returns the current folder where pyadlml saves all datasets to.

    Returns
    -------
    path_to_data_home : str
    """
    import pyadlml
    return pyadlml.DATA_HOME[0]


def load(name):
    """ Loads a python object from the data_home folder if it exists.

    Parameters
    ----------
    param_dict : dict
        A dictionary identifying the object that is to be loaded. The keys and
        values have to exactly match the objects keys and values when it was
        dumped with *dump_in_data_home*.

    Examples
    --------

    Returns
    -------
    X : pd.DataFrame
        Some observations
    y : pd.DataFrame
        Some labels
    """
    # create folder name
    folder_name = hashdict2str({'str': name})
    folder_path = Path(get_data_home()).joinpath(folder_name)

    if not folder_path.exists():
        raise EnvironmentError(f'Nothing is saved in: {folder_path}')

    res = []
    for f in folder_path.glob('*'):
        df = pd.read_parquet(f)
        res.append(df)
   
    return res


def dump(dfs: list, name: str) -> None:
    """ Creates a folder inside the *data_home* and dumps the given dataframe(s) inside.

    Parameters
    ----------
    dfs : list or pd.DataFrame
        The dataframe or dataframes to save
    name : str
        A string used to generate the folder name where the dataframes are stored.

    """
    dfs = dfs if isinstance(dfs, list) else [dfs]

    _ensure_dh_folder_exists()

    # create string representation of dictionary
    folder_name = hashdict2str({'str': name})
    folder_path = Path(get_data_home()).joinpath(folder_name)

    # remove folder if it already exists
    if folder_path.exists():
        shutil.rmtree(folder_path)
        
    _create_folder(folder_path)

    # save all data
    for i, df in enumerate(dfs):
        fp = folder_path.joinpath(f'df_{i}.parquet')
        #df.to_parquet(fp)
        import pyarrow as pa
        import pyarrow.parquet as pq
        table = pa.Table.from_pandas(df)
        pq.write_table(table, fp)


def _create_folder(path_to_folder):
    Path(path_to_folder).mkdir(parents=True, exist_ok=True)


def hashdict2str(param_dict):
    """ creates a unique string for a dictionary 
    Parameters
    ----------
    param_dict : dict
        Contains the attributes of the encoder and the data representations
    Returns
    -------
    folder_name : str
    """
    # sort dictionary after keys for determinism
    param_dict = {k: v for k, v in sorted(param_dict.items(), key=lambda item: item[1])}

    # create string
    param_string = str(param_dict).encode('utf-8')
    folder_name = hashlib.md5(param_string).hexdigest()
    return str(folder_name)


def _delete_data(path_to_folder):
    # make sure only data in home directory are deleted
    #assert '/home/' == path_to_folder[:6]
    shutil.rmtree(path_to_folder)


def _data_2_folder_name(path_to_folder, data_name):
    param_dict = {'dataset': data_name}
    folder_name = hashdict2str(param_dict)
    folder_path = os.path.join(path_to_folder, folder_name)
    return folder_path


def correct_acts(func):
    @functools.wraps(func)
    def wrapper_correction(*args, **kwargs):
        data = func(*args, **kwargs)
        retain_corrections = kwargs['retain_corrections'] if 'retain_corrections' in kwargs else False

        df_acts = data.get_activity_dict()
        for key, df_act in df_acts.items():
            df_act, correction_act = correct_activities(df_act, retain_corrections=retain_corrections)
            data.set_activity_df(df_act, key)
            if retain_corrections:
                data.set_activity_correction(correction_act, key)
        return data
    return wrapper_correction

def correct_devs(func):
    @functools.wraps(func)
    def wrapper_correction(*args, **kwargs):
        data = func(*args, **kwargs)
        retain_corrections = kwargs['retain_corrections'] if 'retain_corrections' in kwargs else False

        df_dev, correction_dev_dict = correct_devices(data.df_devices, retain_correction=retain_corrections)
        data.set_device_df(df_dev)

        if retain_corrections:
            data.set_device_correction(correction_dev_dict[CORRECTION_TS],
                                       correction_dev_dict[CORRECTION_ONOFF_INCONS])

        return data
    return wrapper_correction



def correct_acts_and_devs(func):
    """ Wraps a function that returns a data object with activity, device dataframes
        and applies corrections on to device and activity dataframes.

    """
    @functools.wraps(func)
    def wrapper_correction(*args, **kwargs):
        data = func(*args, **kwargs)
        df_dev = data.df_devices
        df_acts = data.get_activity_dict()

        try:
            retain_corrections = kwargs['retain_corrections']
        except:
            retain_corrections = False

        for key, df_act in df_acts.items():
            df_act, correction_act = correct_activities(df_act, retain_corrections=retain_corrections)
            data.set_activity_df(df_act, key)
            if retain_corrections:
                data.set_activity_correction(correction_act, key)
        df_dev, correction_dev_dict = correct_devices(df_dev, retain_correction=retain_corrections)
        data.set_device_df(df_dev)

        if retain_corrections:
            data.set_device_correction(correction_dev_dict[CORRECTION_TS],
                                       correction_dev_dict[CORRECTION_ONOFF_INCONS])

        return data
    return wrapper_correction



def _move_files_to_parent_folder(path_to_folder):
    """ Moves all files in given folder on level up and deletes the empty directory
    """
    import shutil
    source_dir = Path(path_to_folder)
    for file_name in source_dir.iterdir():
        shutil.move(str(source_dir.joinpath(file_name)), source_dir.parent)

    source_dir.rmdir()


def fetcher(ds_name, func_download, file_name, url, fn_cleaned=None, url_cleaned=None):
    """ The fetcher manages downloading and caching of datasets and their cleaned verions.

    The folder structure for a dataset is
    datahome
    |- dataset_name
        |- original
            |- data files
            |- ...
        cached.joblib
        cleaned.joblib

    Parameters
    ----------
    ds_name : str
        The datasetname. Is also used for the folder
    func_download : function
        The function that downloads the dataset
    file_name : name
        Arguments that are used are passed to the downloader
    """
    FN_CLEANED = 'cleaned.joblib'
    FN_CACHED = 'cached.joblib'
    def decorator_fetcher(func):
        @extract_kwargs
        @functools.wraps(func)
        def wrapper_fetcher(*args, **kwargs):

            # Resolve folders
            data_home = get_data_home()
            _ensure_dh_folder_exists()
            dataset_folder = Path(data_home).joinpath(ds_name)
            original_folder = dataset_folder.joinpath('original')


            # TODO refactor, critical get dataset selecting argument to set prefix for cache and cleaned version
            sel = None
            for a in kwargs.keys():
                if a not in ['cache', 'keep_original', 'retain_corrections',
                            'folder_path', 'load_cleaned']:
                    sel = a
            if sel is not None:
                cached_name = FN_CACHED.split('.')[0] + '_' + kwargs[sel] + '.' + FN_CACHED.split('.')[1]
                cleaned_name = FN_CLEANED.split('.')[0] + '_' + kwargs[sel] + '.' + FN_CLEANED.split('.')[1]
                fp_cached_dataset = dataset_folder.joinpath(cached_name)
                fp_cleaned_dataset = dataset_folder.joinpath(cleaned_name)
            else:
                fp_cached_dataset = dataset_folder.joinpath(FN_CACHED)
                fp_cleaned_dataset = dataset_folder.joinpath(FN_CLEANED)


            try:
                cache = kwargs['cache']
                keep_original = kwargs['keep_original']
                folder_path = kwargs['folder_path']
            except KeyError:
                raise KeyError('The fetch_dataset signature is not set correctly.')
            load_cleaned = kwargs.pop('load_cleaned', False)

            #assert cache and folder_path is None, "Both parameters can not be set at the same time. Please specify only one."

            # Only download data if it was not already fetched
            # and no cached version exists that should be loaded
            # and the load_cleaned flag is not set

            if not (original_folder.exists() or (fp_cached_dataset.is_file() and cache)\
                    or folder_path is not None) and not load_cleaned:
                original_folder.mkdir(parents=True, exist_ok=True)
                func_download(original_folder, file_name, url)
            elif load_cleaned and not fp_cleaned_dataset.exists():
                download_from_mega(dataset_folder, fn_cleaned, url_cleaned, unzip=False)
                Path(dataset_folder).joinpath(fn_cleaned).rename(FN_CLEANED)

            # Load from cached version if available and cache flag is set
            # or load from cleaned version
            # otherwise load from the function
            if fp_cached_dataset.is_file() and cache and not load_cleaned:
                data = joblib.load(fp_cached_dataset)
            elif fp_cleaned_dataset.is_file() and load_cleaned and not cache:
                data = joblib.load(fp_cleaned_dataset)
            else:
                kwargs['folder_path'] = str(original_folder)
                data = func(*args, **kwargs)

            # Save dataset if downloaded and not already cached
            if cache and not fp_cached_dataset.is_file() and not load_cleaned:
                joblib.dump(data, fp_cached_dataset)

            # Clean up data
            # if not cache and fp_cached_dataset.is_file():
            # if not keep_original and original_folder.exists():
            #    _delete_data(original_folder)

            return data
        return wrapper_fetcher
    return decorator_fetcher


def fetch_handler(keep_original, cache, dataset_name,
        mega_filename, mega_url,
        load_func, data_postfix=''):
    """ handles the downloading, loading and caching of a dataset
    Parameters
    ----------
    Returns
    -------
    data : object

    """
    data_home = get_data_home()
    _ensure_dh_folder_exists()
    data_home_dataset = ''.join([data_home, '/', dataset_name])
    cache_data_folder = _data_2_folder_name(data_home, dataset_name)

    # download data    
    if not os.path.isdir(data_home_dataset):
        # download file from mega # TODO make official way available
        download_from_mega(get_data_home(), mega_filename, mega_url)

    # load data
    if data_postfix != '':
        data_name = cache_data_folder + '/' \
            + DATA_DUMP_NAME[:-7] + '_' + data_postfix + '.joblib'
    else:
        data_name = cache_data_folder + '/' + DATA_DUMP_NAME

    if Path(data_name).is_file(): 
        data = joblib.load(data_name) 
    else:
        data = load_func(data_home_dataset + '/')
        if cache:
            _create_folder(cache_data_folder)
            joblib.dump(data, data_name)
            Path(os.path.join(cache_data_folder,dataset_name)).touch()


    # clean up data
    # TODO note that the folder is deleted. For two different subjects
    # caching one and deleting another leads to deletion of the first 
    if not cache and os.path.exists(cache_data_folder):
        _delete_data(cache_data_folder)
    if not keep_original and os.path.exists(data_home_dataset):
        _delete_data(data_home_dataset)

    return data


def clear_data_home():
    """ Delete all content inside the data home folder.
    """
    data_home = get_data_home()
    _delete_data(data_home)
    _create_folder(data_home)


def download_from_mega(path_to_folder, file_name, url, unzip=True):
    """ Downloads dataset from MEGA and extracts it
    Parameters
    ----------
    path_to_folder : PosixPath or str
        The folder where the archive will be extracted to
    file_name : str
        The name of the file to be downloaded
    url : str
        The internet address where mega downloads the file
    unzip : bool, default=True

    """
    file_dp = Path(path_to_folder).joinpath(file_name)
    from mega import Mega

    # Download from mega
    m = Mega()    
    m.download_url(url, dest_path=str(path_to_folder), dest_filename=file_name)

    # Unzip data, remove
    if unzip:
        with zipfile.ZipFile(file_dp, "r") as zip_ref:
            zip_ref.extractall(path_to_folder)
        Path(file_dp).unlink()
        _move_files_to_parent_folder(path_to_folder.joinpath(file_name[:-4]))




import dash
import dash.dcc as dcc
import dash.html as html
import pandas as pd
from dash.exceptions import PreventUpdate
import dash_bootstrap_components as dbc
from pyadlml.dataset.plotly.devices import event_density
from pyadlml.dataset.plotly.layout import DEV_DENS_SLIDER

from pyadlml.dataset._core._dataset import label_data
from pyadlml.dataset._core.activities import _create_activity_df
from pyadlml.dataset._datasets.aras import _create_device_df
from pyadlml.dataset.plotly.activities import *
from pyadlml.dataset.plotly.acts_and_devs import *
from pyadlml.dataset.plotly.devices import bar_count as dev_bar_count, \
    device_iei as dev_iei, boxplot_state
from dash.dependencies import *

from pyadlml.constants import TIME, START_TIME, END_TIME, ACTIVITY, DEVICE, VALUE
from pyadlml.dataset.plotly.layout import acts_vs_devs_layout, devices_layout, \
    activities_layout, acts_n_devs_layout, device_layout_graph_bottom
from pyadlml.dataset.stats.acts_and_devs import contingency_table_states, contingency_table_events
from pyadlml.dataset.util import select_timespan, timestamp_to_num, num_to_timestamp, device_order_by

from dask.delayed import delayed


def _get_plot_height_devs(nr_devs):
    # Determine the plot height and fontsize for activity plots
    if nr_devs < 20:
        return 350
    elif nr_devs < 30:
        return 400
    elif nr_devs < 50:
        return 500
    elif nr_devs < 70:
        return 700
    else:
        return 800


def _get_plot_height_acts(nr_acts):
    if nr_acts < 20:
        return 350
    elif nr_acts < 25:
        return 450
    else:
        return 350




def dashboard(app, name, embedded=False, df_acts=None, df_devs=None, start_time=None, end_time=None):
    """ Creates a dashboard

    Note
    -----
    All parameter may be 'None' if the app is populated during runtime

    """

    # Since performance rendering issues arise with numbers greater than 40000
    # device datapoints make a preselection
    if isinstance(df_acts, pd.DataFrame):
        df_acts = dict(df_acts=df_acts)
    assert isinstance(df_acts, dict)

    max_points = 40000
    if df_devs is not None and len(df_devs) > max_points:
        set_et = df_devs.iat[max_points, df_devs.columns.get_loc(TIME)]
        curr_df_devs, curr_df_acts = select_timespan(
            df_activities=df_acts, df_devices=df_devs, start_time=start_time,
            end_time=set_et, clip_activities=True
        )
    else:
        curr_df_devs = df_devs
        curr_df_acts = df_acts
        set_et = end_time
    
    nr_devs = len(df_devs[DEVICE].unique()) if df_devs is not None else 0
    nr_acts = df_acts.nr_acts() if df_acts is not None else 0

    if start_time is None:
        start_time = min(df_devs[TIME].iloc[0], *[df_acts[k][START_TIME].iloc[0] for k in df_acts.keys()])
        start_time = start_time.floor('D')
    if end_time is None:
        end_time = max(df_devs[TIME].iloc[-1], *[df_acts[k][END_TIME].iloc[-1] for k in df_acts.keys()])
        end_time = end_time.ceil('D')

    # TODO refactor: ugly as hell defined in 76
    if df_devs is None or len(df_devs) <= max_points:
        set_et = end_time

    plot_height_devs = _get_plot_height_devs(nr_devs)
    plot_height_acts = _get_plot_height_acts(nr_acts)

    # Get Layout
    layout_activities = [activities_layout(df, k, plot_height_acts) for k, df in curr_df_acts.items()]
    layout_devices = devices_layout(curr_df_devs, True, plot_height_devs)
    layout_acts_vs_devs = acts_vs_devs_layout(
        list(curr_df_acts.values())[0],
        curr_df_devs,
        True,
        plot_height_acts
    )

    activity_tabs = [dbc.Tab(l, label=f'Activities {act_name}', tab_id=f'tab-acts-{act_name}') for act_name, l in zip(curr_df_acts.keys(), layout_activities)]
    content = [
            dcc.Input(id='act_assist_path', type='hidden', value='filler text'),
            dcc.Input(id='subject_names', type='hidden', value='filler text'),
            html.H2(children=f'Dashboard: {name}'),
            acts_n_devs_layout(df_acts, df_devs, start_time, end_time,
                               set_et, plot_height_devs+50),
            html.Br(),
        html.Div([
            dbc.Tabs(
                [*activity_tabs,
                 dbc.Tab(layout_devices, label='Devices', tab_id='tab-devs'),
                 dbc.Tab(layout_acts_vs_devs, label='Activities ~ Devices', tab_id='tab-acts_vs_devs'),
                 ], id='tabs', active_tab='tab-acts',
            ),
            html.Div(id="content"),
        ]),
        # Store intermediate values of hard to compute values
        # dcc.Store(id='current-df-acts'),
    ]

    if embedded:
        layout = dbc.Container(
            children=content, 
            style={'width': 1000, 'margin': 'auto'}
        )
    else:
        layout = dbc.Container(
            children=content,
            style={'margin': 'auto'}
        )
    app.layout = layout


    create_callbacks(app, df_acts, df_devs, start_time, end_time, plot_height_acts, plot_height_devs)

def _sel_avd_event_click(avd_event_select, curr_df_acts, curr_df_devs):
    """Select devices and activities from click data of the event contingency table"""
    # Select devices
    a = avd_event_select['points'][0]['y']
    d = avd_event_select['points'][0]['x']
    df_tmp = curr_df_devs[curr_df_devs[DEVICE] == d]
    df_tmp = label_data(df_tmp, curr_df_acts)
    df_tmp = df_tmp[df_tmp[ACTIVITY] == a]
    sel_devices = df_tmp[[TIME, DEVICE, VALUE]]

    # Select activities
    sel_activities = _create_activity_df()
    for _, sel_dev in sel_devices.iterrows():
        mask = (curr_df_acts[START_TIME] < sel_dev[TIME]) \
               & (sel_dev[TIME] < curr_df_acts[END_TIME])
        tmp = curr_df_acts[mask]
        sel_activities = pd.concat([sel_activities, tmp])
    sel_activities = sel_activities.drop_duplicates()
    return sel_devices, sel_activities


def _sel_point_to_activities(selection):
    """ Gets the selected activities from the boxplot or violin-plot and
        asdf
    """
    user_sel_activities = []
    points = selection['points']
    for usa in points:
        user_sel_activities.append([usa['customdata'][0], usa['customdata'][1], usa['y']])

    df_usa = pd.DataFrame(user_sel_activities, columns=[START_TIME, END_TIME, ACTIVITY])
    df_usa[START_TIME] = pd.to_datetime(df_usa[START_TIME])
    df_usa[END_TIME] = pd.to_datetime(df_usa[END_TIME])
    return df_usa

def _sel_act_bar(selection, df_acts):
    """ Retrieve the activity from a click on an activity bar in a
        bar plot e.g count, cummulative and filter the activity dataframe

    """
    dev = selection['points'][0]['y']
    return df_acts[df_acts[ACTIVITY] == dev]


def _sel_dev_bar(selection, df_devs):
    """ Retrieve the device from a click on the device bar the event count
        graphand filter the device dataframe

    """
    dev = selection['points'][0]['y']
    return df_devs[df_devs[DEVICE] == dev]

def _sel_dev_fraction(selection, df_devs):
    """ Retrieve the device from a click on the device bar the event count
        graphand filter the device dataframe

    """
    dev = selection['points'][0]['y']
    cat = selection['points'][0]['customdata'][1]
    mask = (df_devs[DEVICE] == dev) & (df_devs[VALUE] == cat)
    return df_devs[mask].copy()

def _sel_dev_bp_selection(dev_boxplot_select, df_devs):
    """ Gets point data from the state boxplot and creates
        an device-dataframe
    """
    points = dev_boxplot_select['points']
    sd = pd.DataFrame(columns=[TIME, END_TIME, DEVICE, VALUE])
    for p in points:
        duration = pd.Timedelta(p['customdata'][0])
        time = p['customdata'][1]
        dev = p['y']
        mask = (df_devs[DEVICE] == dev) & (df_devs[TIME] == time)
        df = df_devs[mask]
        df[END_TIME] = df[TIME] + duration
        sd = pd.concat([sd, df])

    return sd


def _sel_dev_iei(dev_iei_select, df_devs, scale, fig):

    per_device = not (len(fig['data']) == 1)
    df = df_devs.copy()
    df_res = pd.DataFrame(columns=[TIME, DEVICE, VALUE])
    df['diff'] = df[TIME].shift(-1) - df[TIME]
    size = fig['data'][0]['xbins']['size']

    for p in dev_iei_select['points']:
        left, right = p['x'] - size/2, p['x'] + size/2
        if scale == 'log':
            left, right = np.exp(left), np.exp(right)
        mask = (pd.to_timedelta(left, unit='s') <= df['diff']) \
               & (df['diff'] <= pd.to_timedelta(right, unit='s'))
        if per_device:
            mask = mask & (df[DEVICE] == p['customdata'])

        count = sum(mask)
        sel_count = len(p['pointNumbers'])
        assert count == sel_count, f'Lookup error: The selected device number {sel_count} != {count}'
        df_res = pd.concat([df_res, df_devs[mask]])

    return df_res


def _sel_dev_density(selection, df_devs, dt):

    from pyadlml.dataset.plotly.layout import DEV_DENS_SLIDER
    df = df_devs.copy().set_index(TIME, drop=False)
    dt = pd.Timedelta(DEV_DENS_SLIDER[dt])
    df_res = pd.DataFrame(columns=[TIME, DEVICE, VALUE])

    for p in selection['points']:
        dev = p['y']
        count = p['z']
        mid = pd.Timestamp(p['x'])
        tmp = df[df[DEVICE] == dev].between_time(
                    start_time=(mid - dt/2).time(),
                    end_time=(mid + dt/2).time())\
              .reset_index(drop=True)

        assert count == len(tmp.index)
        df_res = pd.concat([df_res, tmp])

    if df_res.empty:
        return None
    else:
        return df_res


def _sel_act_trans_sel(act_trans_select, df_acts):
    """ Gets point data from the state boxplot and creates
        an device-dataframe
    """
    df = df_acts.copy()\
        .sort_values(by=START_TIME)\
        .reset_index(drop=True)
    act1 = act_trans_select['points'][0]['x']
    act2 = act_trans_select['points'][0]['y']
    df[ACTIVITY + '2'] = df[ACTIVITY].shift(-1)
    idx = df[(df[ACTIVITY] == act1) & (df[ACTIVITY + '2'] == act2)].index.values
    df4 = df.iloc[[*idx, *(idx+1)]]
    df4 = df4[[START_TIME, END_TIME, ACTIVITY]]\
            .sort_values(by=START_TIME)\
            .reset_index(drop=True)
    return df4


def _initialize_toggle_callbacks(app):
    @app.callback(
        Output("clps-act-transition", "is_open"),
        [Input("clps-act-transition-button", "n_clicks")],
        [State("clps-act-transition", "is_open")],
    )
    def toggle_collapse(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-avd-state", "is_open"),
        [Input("clps-avd-state-button", "n_clicks")],
        [State("clps-avd-state", "is_open")],
    )
    def toggle_collapse(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-avd-event", "is_open"),
        [Input("clps-avd-event-button", "n_clicks")],
        [State("clps-avd-event", "is_open")],
    )
    def toggle_collapse(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-acts-n-devs", "is_open"),
        [Input("clps-acts-n-devs-button", "n_clicks")],
        [State("clps-acts-n-devs", "is_open")],
    )
    def toggle_collapse(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-act-boxplot", "is_open"),
        [Input("clps-act-boxplot-button", "n_clicks")],
        [State("clps-act-boxplot", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-act-bar", "is_open"),
        [Input("clps-act-bar-button", "n_clicks")],
        [State("clps-act-bar", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-dev-iei", "is_open"),
        [Input("clps-dev-iei-button", "n_clicks")],
        [State("clps-dev-iei", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-dev-bar", "is_open"),
        [Input("clps-dev-bar-button", "n_clicks")],
        [State("clps-dev-bar", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-dev-bp", "is_open"),
        [Input("clps-dev-bp-button", "n_clicks")],
        [State("clps-dev-bp", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)

    @app.callback(
        Output("clps-dev-density", "is_open"),
        [Input("clps-dev-density-button", "n_clicks")],
        [State("clps-dev-density", "is_open")],
    )
    def toggle_collapse_act_bp(n, is_open):
        return bool(n) ^ bool(is_open)


def create_callbacks(app, df_acts, df_devs, start_time, end_time, plt_height_acts, plt_height_devs):

    def gen_trigger(id, val):
        return html.Div(id=id, style=dict(display="none"), **{"data-value": val})

    _initialize_toggle_callbacks(app)

    @app.callback(
        Output("and_clipboard", "content"),
        Input("graph-acts_n_devs", "clickData"),
        State("and_dev-type", "value")
    )
    def element_to_clipboard(tmp, dev_type):
        if tmp is None:
            raise PreventUpdate
        dp = tmp['points'][0]
        is_activity = (dp['y'] == 'Activity')
        is_dev_state = not is_activity and dev_type == 'state'
        is_dev_event = not is_activity and dev_type == 'event'
        if is_activity:
            start_time = dp['base']
            end_time = dp['x']
            s = f'\'{start_time}\',\'{end_time}\',#activity'
        elif is_dev_state:
            start_time = dp['base']
            end_time = dp['customdata'][1]
            dev = dp['y']
            val = dp['customdata'][2]
            s = f'\'{start_time}\',\'{end_time}\',\'{dev}\',\'{val}\''
        elif is_dev_event:
            time = dp['x']
            dev = dp['y']
            val = dp['customdata']
            s = f'\'{time}\',\'{dev}\',\'{val}\''
        else:
            raise NotImplementedError
        return s

    @app.callback(
        output=[
            Output('graph-acts_n_devs', 'figure'),
            Output('and_act-order', 'data'),
            Output('and_dev-order', 'data'),
            Output('and_reset_sel', 'disabled'),
            Output('act-trigger', 'children'),
            Output('avd-trigger', 'children'),
            Output('dev-trigger', 'children'),
            Output('act-curr-sel', 'children'),
            Output('dev-curr-sel', 'children'),
            Output('act-curr-sel-store', 'data'),
            Output('dev-curr-sel-store', 'data'),
        ],
        inputs=[
            # Input methods
            Input('range-slider', 'value'),
            Input('select-activities', 'value'),
            Input('select-devices', 'value'),
            Input('and_act-order-trigger', 'value'),
            Input('and_dev-order-trigger', 'value'),
            Input('and_dev-type', 'value'),

            # Selected activities
            Input('graph-boxplot', 'selectedData'),
            Input('graph-transition', 'clickData'),
            Input('graph-bar', 'clickData'),

            # Selected devices
            Input('devs_graph-bar', 'clickData'),
            Input('devs_graph-boxplot', 'selectedData'),
            Input('devs_graph-iei', 'selectedData'),
            Input('devs_graph-fraction', 'clickData'),

            Input('devs_graph-density', 'clickData'),
            Input('avd_graph-event-contingency', 'clickData'),
            Input('and_reset_sel', 'n_clicks'),
        ],
        state=[State('devs_dens-slider', 'value'),
            State('devs_iei-scale', 'value'),
            State('devs_graph-iei', 'figure'),
            State('and_act-order', 'data'),
            State('and_dev-order', 'data'),
            State('act-curr-sel', 'children'),
            State('dev-curr-sel', 'children'),
            State('act-curr-sel-store', 'data'), 
            State('dev-curr-sel-store', 'data'), 
        ]
    )
    def update_acts_n_devs(rng, sel_activities, sel_devices, act_order_trigger,
                           dev_order_trigger, dev_type_trigger,
                           act_boxplot_select, act_trans_select, act_bar_select,
                           dev_bar_select, dev_boxplot_select, dev_iei_select,
                           dev_fraction_select, 
                           dev_density_select, avd_event_select, reset_sel,

                           dev_density_dt, dev_iei_scale, dev_iei_fig,
                           act_order, dev_order, act_curr_sel, dev_curr_sel,
                           act_curr_sel_store, dev_curr_sel_store, 
                           ):
        ctx = dash.callback_context

        if _get_trigger_value(ctx) is None:
            raise PreventUpdate

        # Catch block if no act-order or dev-order is initialized
        try:
            act_order = json.loads(act_order)
        except TypeError:
            act_order = None
        try:
            dev_order = json.loads(dev_order)
        except TypeError:
            dev_order = None
        try:
            df_act_curr_sel = pd.read_json(act_curr_sel_store)
        except:
            df_act_curr_sel = None
        try:
            df_dev_curr_sel = pd.read_json(dev_curr_sel_store)
        except:
            df_dev_curr_sel = None

        origin_act_trans_selection = _is_trigger(ctx, 'graph-transition')
        origin_act_bp_selection = _is_trigger(ctx, 'graph-boxplot')
        origin_act_bar_selection = _is_trigger(ctx, 'graph-bar')
        origin_and_event_selection = _is_trigger(ctx, 'avd_graph-event-contingency')
        origin_dev_bp_selection = _is_trigger(ctx, 'devs_graph-boxplot')
        origin_dev_iei_selection = _is_trigger(ctx, 'devs_graph-iei')
        origin_dev_density_selection = _is_trigger(ctx, 'devs_graph-density')
        origin_dev_bar_selection = _is_trigger(ctx, 'devs_graph-bar')
        origin_dev_fraction_selection = _is_trigger(ctx, 'devs_graph-fraction')

        data_selection_changed = _is_trigger(ctx, 'range-slider') or _is_trigger(ctx, 'select-devices') \
                                 or _is_trigger(ctx, 'select-activities')

        user_sel_activities = origin_act_trans_selection \
                              or origin_act_bp_selection \
                              or origin_and_event_selection \
                              or origin_act_bar_selection
        user_sel_devices = origin_and_event_selection \
                           or origin_dev_bp_selection \
                           or origin_dev_iei_selection \
                           or origin_dev_density_selection \
                           or origin_dev_bar_selection \
                           or origin_dev_fraction_selection

        reset_selection = _is_trigger(ctx, 'and_reset_sel') or data_selection_changed

        # Create start- and end-timestamps from range slider
        st = num_to_timestamp(rng[0], start_time=start_time, end_time=end_time)
        et = num_to_timestamp(rng[1], start_time=start_time, end_time=end_time)
        curr_df_devs, curr_df_acts = select_timespan(df_activities=df_acts, df_devices=df_devs,
                                                     start_time=st, end_time=et, clip_activities=True)

        # Filter selected devices and activities
        curr_df_devs = curr_df_devs[curr_df_devs[DEVICE].isin(sel_devices)]
        curr_df_acts = curr_df_acts[curr_df_acts[ACTIVITY].isin(sel_activities)]
        # Filter for acts-n-devs plot
        df_devs1 = df_devs[df_devs[DEVICE].isin(sel_devices)]
        df_acts1 = df_acts[df_acts[ACTIVITY].isin(sel_activities)]

        devs_usel_state = None
        # Create activity and device order if needed
        if _is_trigger(ctx, 'and_dev-order-trigger') or dev_order is None:
            dev_order = device_order_by(curr_df_devs, dev_order_trigger)
        if _is_trigger(ctx, 'and_act-order-trigger') or act_order is None:
            act_order = activity_order_by(curr_df_acts, act_order_trigger)


        # Create dataframes with user selection
        if (_is_trigger(ctx, 'graph-transition') or act_curr_sel == 'graph-transition')\
            and not reset_selection:
            df_act_curr_sel = _sel_act_trans_sel(act_trans_select,
                                                        df_acts=curr_df_acts)
            act_curr_sel = 'graph-transition'
        if (_is_trigger(ctx, 'graph-boxplot') or act_curr_sel == 'graph-boxplot')\
            and not reset_selection:
            df_act_curr_sel = _sel_point_to_activities(act_boxplot_select)
            act_curr_sel = 'graph-boxplot'

        if (_is_trigger(ctx, 'graph-bar') or act_curr_sel == 'graph-bar')\
            and not reset_selection:
            df_act_curr_sel = _sel_act_bar(act_bar_select, df_acts=curr_df_acts)
            act_curr_sel = 'graph-bar'

        #if update_devices:

        if (_is_trigger(ctx, 'devs_graph-bar') or dev_curr_sel == 'devs_graph-bar')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_bar(dev_bar_select, df_devs=curr_df_devs)
            dev_curr_sel = 'devs_graph-bar'
            devs_usel_state = False

        if (_is_trigger(ctx, 'devs_graph-fraction') or dev_curr_sel == 'devs_graph-fraction')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_fraction(dev_fraction_select, df_devs=curr_df_devs)
            dev_curr_sel = 'devs_graph-fraction'
            devs_usel_state = True

        if (_is_trigger(ctx, 'devs_graph-boxplot') or dev_curr_sel == 'devs_graph-boxplot')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_bp_selection(dev_boxplot_select, curr_df_devs)
            dev_curr_sel = 'devs_graph-boxplot'
            devs_usel_state = True

        if (_is_trigger(ctx, 'devs_graph-iei') or dev_curr_sel == 'devs_graph-iei')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_iei(dev_iei_select, curr_df_devs,
                                           dev_iei_scale, dev_iei_fig)
            dev_curr_sel = 'devs_graph-iei'
            devs_usel_state = False

        if (_is_trigger(ctx, 'devs_graph-density') or dev_curr_sel == 'devs_graph-density')\
            and not reset_selection:
            df_dev_curr_sel = _sel_dev_density(dev_density_select, curr_df_devs, dev_density_dt)
            dev_curr_sel = 'devs_graph-density'
            devs_usel_state = False

        if _is_trigger(ctx, 'avd_graph-event-contingency') and not reset_selection:
            df_dev_curr_sel, df_act_curr_sel = _sel_avd_event_click(avd_event_select,
                                                               curr_df_acts,
                                                               curr_df_devs)
            act_curr_sel = 'avd_graph-event-contingency'
            dev_curr_sel = 'avd_graph-event-contingency'
            devs_usel_state = False

        if reset_selection:
            df_act_curr_sel = None
            df_dev_curr_sel = None
            act_curr_sel = ''
            dev_curr_sel = ''

        states = (dev_type_trigger == 'state')
        fig_and = activities_and_devices(df_acts1, df_devs1, st=st, et=et,
                                         states=states, act_order=act_order, dev_order=dev_order,
                                         df_acts_usel=df_act_curr_sel, df_devs_usel=df_dev_curr_sel,
                                         devs_usel_state=devs_usel_state, height=plt_height_devs+50
                                         )

        # Determine reset selection button values
        if user_sel_activities or user_sel_devices:
            disable_reset = False
        else:
            disable_reset = True

        # Determine if and what updates for activities, device or a~d section are needed
        if data_selection_changed:
            act_update, avd_update, dev_update = 'new_data', 'new_data', 'new_data'
            disable_reset = True
        elif reset_selection:
            act_update, dev_update = "reset_sel", "reset_sel"
            avd_update = dash.no_update
        else:
            if origin_act_bp_selection:
                act_update = 'reset_sel_trans'
            elif origin_act_trans_selection:
                act_update = 'reset_sel_bp'
            elif origin_and_event_selection:
                act_update = 'reset_sel'
            else:
                act_update = dash.no_update

            if origin_dev_density_selection:
                dev_update = 'dens_clicked'
            elif origin_dev_iei_selection:
                dev_update = 'iei_selected'
            elif origin_dev_bp_selection:
                dev_update = 'bp_selected'
            else:
                dev_update = dash.no_update

            # TODO, refactor avd is never updates since it is not necessary
            avd_update = dash.no_update
        try:
            act_curr_sel_store = df_act_curr_sel.to_json(date_unit="ns")
        except:
            act_curr_sel_store = ''
        try:
            dev_curr_sel_store = df_dev_curr_sel.to_json(date_unit="ns"),
        except:
            dev_curr_sel_store = ''

        return fig_and, act_order, dev_order, disable_reset, \
               act_update, avd_update, dev_update, act_curr_sel, dev_curr_sel,\
               act_curr_sel_store, dev_curr_sel_store


    @app.callback(
       output=[
            Output('devs_graph-iei', 'figure'),
            Output('devs_graph-density', 'figure'),
            Output('devs_graph-bar', 'figure'),
            Output('devs_graph-fraction', 'figure'),
            Output('devs_graph-boxplot', 'figure'),
            Output('devs_order', 'data'),
        ],
        inputs=[
            Input('dev-trigger', 'children'),
            Input('range-slider', 'value'),
            Input('select-devices', 'value'),
            Input('tabs', 'active_tab'),
            Input('devs_bar-scale', 'value'),
            Input('devs_bar-order', 'value'),

            Input('devs_iei-scale', 'value'),
            Input('devs_iei-per-device', 'on'),
            Input('devs_bp-scale', 'value'),
            Input('devs_bp-binary-state', 'value'),
            Input('devs_dens-slider', 'value'),
            Input('devs_dens-scale', 'value'),
        ],
        state=[
            State('devs_graph-iei', 'figure'),
            State('devs_graph-density', 'figure'),
            State('devs_graph-bar', 'figure'),
            State('devs_graph-fraction', 'figure'),
            State('devs_graph-boxplot', 'figure'),
            State('devs_order', 'data'),
        ]
    )
    def update_dev_tab(dev_trigger, rng, sel_devices,
                       active_tab, bar_scale, bar_order, iei_scale, iei_per_dev,
                       bp_scale, bp_binary_state, dens_slider, dens_scale,
                       fig_iei, fig_dens, fig_left, fig_frac, fig_bp, dev_order, 
                       ):
        ctx = dash.callback_context
        if _get_trigger_value(ctx) is None or active_tab != 'tab-devs':
            raise PreventUpdate

        try:
            dev_order = json.loads(dev_order)
        except TypeError:
            dev_order = None

        # Filter selected timeframe, activities and devices
        st = num_to_timestamp(rng[0], start_time=start_time, end_time=end_time)
        et = num_to_timestamp(rng[1], start_time=start_time, end_time=end_time)
        curr_df_devs = select_timespan(df_devices=df_devs, start_time=st, end_time=et,
                                       clip_activities=True)
        curr_df_devs = curr_df_devs[curr_df_devs[DEVICE].isin(sel_devices)]
        from pyadlml.dataset.plotly.devices import fraction as dev_fraction,\
                                       bar_count

        is_trigger_bar_drop = _is_trigger(ctx, 'devs_bar-drop')
        is_trigger_bar_scale = _is_trigger(ctx, 'devs_bar-scale')
        is_trigger_bp_scale = _is_trigger(ctx, 'devs_bp-scale')
        is_trigger_bp_state = _is_trigger(ctx, 'devs_bp-binary-state')
        is_trigger_iei_per_device = _is_trigger(ctx, 'devs_iei-per-device')
        is_trigger_iei_scale = _is_trigger(ctx, 'devs_iei-scale')
        is_trigger_dens_scale = _is_trigger(ctx, 'devs_dens-scale')
        is_trigger_dens_slider = _is_trigger(ctx, 'devs_dens-slider')
        is_trigger_sel_dev = _is_trigger(ctx, 'select-devices')
        is_trigger_dev_update = _is_trigger(ctx, 'dev-trigger') or is_trigger_sel_dev
        order_update = _is_trigger(ctx, 'devs_bar-order')
        new_data_signal = (is_trigger_dev_update and dev_trigger == 'new_data')

        update_bar = is_trigger_bar_drop or order_update or is_trigger_bar_scale \
                    or new_data_signal
        update_bp = is_trigger_bp_scale or order_update or is_trigger_bp_state \
                    or (is_trigger_dev_update and dev_trigger == 'dens_clicked') \
                    or (is_trigger_dev_update and dev_trigger == 'iei_selected') \
                    or (is_trigger_dev_update and dev_trigger == 'reset_sel') \
                    or new_data_signal
        update_iei = is_trigger_iei_scale or is_trigger_iei_per_device \
                    or new_data_signal
        update_frac = order_update or new_data_signal
        update_density = is_trigger_dens_slider or order_update \
                    or new_data_signal or is_trigger_dens_scale

        if order_update or dev_order is None or is_trigger_dev_update or is_trigger_sel_dev:
            dev_order = device_order_by(curr_df_devs, rule=bar_order)



        def f_update_frac(fig_frac, cd_dev, order, uf):
            return dev_fraction(cd_dev, order=order, height=plt_height_devs) if uf else fig_frac
        def f_update_bar(ub, fig_left, cdev, scale, order):
            return bar_count(cdev, scale=scale, order=order, height=plt_height_devs) if ub else fig_left
        def f_update_bp(u_bp, fig_bp, cd_dev, order, scale, binary_state):
            if u_bp:
                return boxplot_state(cd_dev, scale=scale, order=order,
                                     binary_state=binary_state, height=plt_height_devs)
            else:
                return fig_bp
        def f_update_density(ud, fig_dens, cd_dev, dens_slider, scale, order):
            if ud:
                return event_density(cd_dev, dt=DEV_DENS_SLIDER[dens_slider], show_colorbar=False,
                                     scale=scale, order=order, height=plt_height_devs)
            else:
                return fig_dens
        def f_update_iei(u_iei, fig_iei, cd_dev, scale, per_dev):
            return dev_iei(cd_dev, scale=scale, per_device=per_dev, height=plt_height_devs) if u_iei else fig_iei

        import dask
        figs = [
            dask.delayed(f_update_iei)(update_iei, fig_iei, curr_df_devs, iei_scale, iei_per_dev),
            dask.delayed(f_update_density)(update_density, fig_dens, curr_df_devs,
                                           dens_slider, dens_scale, dev_order),
            dask.delayed(f_update_bar)(update_bar, fig_left, curr_df_devs, bar_scale, dev_order),
            dask.delayed(f_update_frac)(fig_frac, curr_df_devs, dev_order, update_frac),
            dask.delayed(f_update_bp)(update_bp, fig_bp, curr_df_devs, dev_order, bp_scale,
                                      bp_binary_state),
        ]
        figs = dask.compute(figs)[0]

        return figs[0], figs[1], figs[2], figs[3], figs[4], json.dumps(list(dev_order))


    @app.callback(
        output=[
            Output('graph-bar', 'figure'),
            Output('graph-boxplot', 'figure'),
            Output('graph-density', 'figure'),
            Output('graph-transition', 'figure'),
            Output('acts_activity-order', 'data'),
            Output('acts_density-data', 'data')
        ],
        inputs=[
            Input('tabs', 'active_tab'),
            Input("act-trigger", 'children'),
            Input('range-slider', 'value'),
            Input('select-activities', 'value'),
            Input('acts_bp-drop', 'value'),
            Input('acts_bar-drop', 'value'),
            Input('acts_bp-scale', 'value'),
            Input('acts_bar-scale', 'value'),
            Input('acts_trans-scale', 'value'),
            Input('acts_sort', 'value'),
        ],
        state=[
            State('graph-bar', 'figure'),
            State('graph-boxplot', 'figure'),
            State('graph-density', 'figure'),
            State('graph-transition', 'figure'),
            State('acts_activity-order', 'data'),
            State('acts_density-data', 'data'),
        ]

    )
    def update_activity_tab(active_tab, act_trigger, rng, sel_activities,
                               drop_box: str, drop_bar: str, scale_boxplot: str,
                               scale_bar: str, scale_trans: str, act_order_trigger,
                               fig_bar, fig_bp, fig_dens, fig_trans, act_order, act_density,
        ):

        ctx = dash.callback_context
        if _get_trigger_value(ctx) is None or active_tab != 'tab-acts':
            raise PreventUpdate

        try:
            act_order = json.loads(act_order)
        except TypeError:
            # happens if no act-order is initialized
            act_order = None
        try:
            act_density = pd.read_json(act_density)
        except:
            act_density = None

        # Filter selected timeframe, activities and devices
        st = num_to_timestamp(rng[0], start_time=start_time, end_time=end_time)
        et = num_to_timestamp(rng[1], start_time=start_time, end_time=end_time)
        curr_df_acts = select_timespan(df_activities=df_acts, start_time=st, end_time=et,
                                       clip_activities=True)
        curr_df_acts = curr_df_acts[curr_df_acts[ACTIVITY].isin(sel_activities)]
        # TODO refactor, where is the y_label column coming from
        curr_df_acts = curr_df_acts[[START_TIME, END_TIME, ACTIVITY]]

        # Get update type
        is_trigger_bar_drop = _is_trigger(ctx, 'acts_bar-drop')
        is_trigger_bar_scale = _is_trigger(ctx, 'acts_bar-scale')
        is_trigger_sort = _is_trigger(ctx, 'acts_sort')
        is_trigger_range = _is_trigger(ctx, 'range-slider')
        is_trigger_bp_drop = _is_trigger(ctx, 'acts_bp-drop')
        is_trigger_bp_scale = _is_trigger(ctx, 'acts_bp-scale')
        is_trigger_trans_scale = _is_trigger(ctx, 'acts_trans-scale')

        data_update = is_trigger_range or _is_trigger(ctx, 'select-activities')
        # Determine the intent, when the trigger was the avd plot
        signal_reset_bp = _is_trigger(ctx, 'act-trigger') and act_trigger == 'reset_sel_bp'
        signal_reset_trans = _is_trigger(ctx, 'act-trigger') and act_trigger == 'reset_sel_trans'
        signal_reset_all = _is_trigger(ctx, 'act-trigger') and act_trigger == 'reset_sel'

        order_update = ((is_trigger_sort or (is_trigger_bar_drop and act_order_trigger == 'value')) \
                       and not signal_reset_all) \
                       or act_order is None

        bp_update = is_trigger_bp_drop or is_trigger_bp_scale \
                    or data_update or order_update or signal_reset_all\
                    or signal_reset_bp
        bar_update = is_trigger_bar_drop or order_update or data_update \
                     or is_trigger_bar_scale
        trans_update = data_update or order_update or is_trigger_trans_scale \
                       or signal_reset_all

        # If the activity-order is changed or the bar plot is changed
        # and would change the order
        if order_update:
            if act_order_trigger == 'value' or is_trigger_bar_drop:
                act_order_trigger = 'duration' if drop_bar == 'cum' else 'count'
            act_order = activity_order_by(curr_df_acts, act_order_trigger)

        # Update activity bars
        if bar_update:
            if drop_bar == 'count':
                fig_bar = bar_count(curr_df_acts, order=act_order, scale=scale_bar,
                                    height=plt_height_acts)
            else:
                fig_bar = bar_cum(curr_df_acts, order=act_order, scale=scale_bar,
                                  height=plt_height_acts)

        # Update log for boxplot
        if bp_update:
            #if _get_trigger_value(ctx) == 'vp':
            #    fig_bp = violin_duration(curr_df_acts, order=act_order, scale=scale_boxplot)
            #else:
            #    fig_bp = boxplot_duration(curr_df_acts, order=act_order, scale=scale_boxplot)
            fig_bp = boxplot_duration(curr_df_acts, order=act_order, scale=scale_boxplot,
                                      height=plt_height_acts)

        # Only update the act_density matrix if it is
        if data_update or act_density is None:
            act_density = activities_dist(curr_df_acts.copy(), n=1000, dt=None)
        if order_update or data_update:
            fig_dens = density(df_density=act_density, order=act_order, height=plt_height_acts)

        if trans_update:
            fig_trans = heatmap_transitions(curr_df_acts, order=act_order, scale=scale_trans,
                                            height=plt_height_acts)

        return fig_bar, fig_bp, fig_dens, fig_trans, json.dumps(list(act_order)), act_density.to_json()

    @app.callback(
        output=[
            Output('avd_graph-event-contingency', 'figure'),
            Output('avd_graph-state-contingency', 'figure'),
            Output('loading-output', 'children'),
            Output('avd-update', 'children'),
            Output('avd_state-contingency', 'data'),
            Output('avd_event-contingency', 'data'),
            Output('avd_activity-order', 'data'),
            Output('avd_device-order', 'data'),
        ],
        inputs=[
            Input('tabs', 'active_tab'),
            Input('avd-trigger', 'children'),
            Input('avd-update', 'children'),
            Input('avd_state-scale', 'value'),
            Input('avd_event-scale', 'value'),
            Input('avd_act-order-trigger', 'value'),
            Input('avd_dev-order-trigger', 'value'),
        ],
        state=[
            State('range-slider', 'value'),
            State('select-activities', 'value'),
            State('select-devices', 'value'),
            State('avd_state-contingency', 'data'),
            State('avd_event-contingency', 'data'),
            State('avd_activity-order', 'data'),
            State('avd_device-order', 'data'),
        ]
    )
    def update_acts_vs_devs_tab(active_tab, trigger, update,  state_scale,  event_scale,
                                 act_order_trigger, dev_order_trigger,
                                rng, sel_activities, sel_devices,
                                state_cont, event_cont, act_order, dev_order,
                                
        ):
        ctx = dash.callback_context

        if _get_trigger_value(ctx) is None or active_tab != 'tab-acts_vs_devs':
            raise PreventUpdate

        try:
            df_con_states = pd.read_json(state_cont)
            df_con_states = df_con_states.astype('timedelta64[ns]')
            df_con_events = pd.read_json(event_cont)
        except ValueError:
            df_con_states = None
            df_con_events = None
        except BaseException as err:
            print(f"Unexpected {err=}, {type(err)=}")
            raise

        try:
            act_order = json.loads(act_order)
            dev_order = json.loads(dev_order)
        except TypeError:
            act_order = None
            dev_order = None


        # Filter selected timeframe, activities and devices
        st = num_to_timestamp(rng[0], start_time=start_time, end_time=end_time)
        et = num_to_timestamp(rng[1], start_time=start_time, end_time=end_time)
        curr_df_devs, curr_df_acts = select_timespan(df_activities=df_acts, df_devices=df_devs,
                                                     start_time=st, end_time=et, clip_activities=True)
        curr_df_acts = curr_df_acts[curr_df_acts[ACTIVITY].isin(sel_activities)]
        curr_df_devs = curr_df_devs[curr_df_devs[DEVICE].isin(sel_devices)]
        # TODO refactor, where is the y_label column coming from
        curr_df_acts = curr_df_acts[[START_TIME, END_TIME, ACTIVITY]]

        # If the activity-order is changed or the bar plot is changed
        # and would change the order
        is_data_update = _is_trigger(ctx, 'new_data')
        is_trigger_act_order = _is_trigger(ctx, 'avd_act-order-trigger')
        is_trigger_dev_order = _is_trigger(ctx, 'avd_dev-order-trigger')
        order_update = is_trigger_act_order or is_trigger_dev_order\
                       or act_order is None or dev_order is None

        if order_update:
            act_order = activity_order_by(curr_df_acts, act_order_trigger)
            dev_order = device_order_by(curr_df_devs, dev_order_trigger)


        # If data has changed the contingency tables have to be recomputed
        if is_data_update or df_con_states is None or df_con_events is None:
            # Recompute state contingency
            df_con_states = contingency_table_states(curr_df_devs, curr_df_acts,
                                                     distributed=True)
            df_con_events = contingency_table_events(curr_df_devs, curr_df_acts)
            update_data = True
        else:
            update_data = False

        if active_tab != 'tab-acts_vs_devs' and is_data_update:
            return dash.no_update, dash.no_update, dash.no_update, 'updated', \
                   df_con_states.to_json(), df_con_events.to_json(), \
                   json.dumps(list(act_order)), json.dumps(list(dev_order))
        elif active_tab != 'tab-acts_vs_devs':
            raise PreventUpdate

        # Create figures
        fig_adec = contingency_events(
                    con_tab=df_con_events, scale=event_scale,
                    act_order=act_order, dev_order=dev_order
        )
        fig_adsc = contingency_states(
                    df_devs=curr_df_devs,   # for event order
                    con_tab=df_con_states, scale=state_scale,
                    act_order=act_order, dev_order=dev_order
        )

        if update_data:
            state_dump, event_dump = df_con_states.to_json(date_unit="ns"), df_con_events.to_json()
        else:
            state_dump, event_dump = dash.no_update, dash.no_update

        return fig_adec, fig_adsc, None, 'needs update', state_dump, event_dump, \
               json.dumps(list(act_order)), json.dumps(list(dev_order))


    #@app.callback(
    #    Output('graph-acts_n_devs', 'figure'),
    #    Input('and_dev-type', 'value'),
    #    State('act_assist_path', 'value'),
    #    State('subject_names', 'value'),
    #)
    #def update_acts_n_devs(dev_type_trigger, act_assist_path, subject_names):
    #    data = load_act_assist(act_assist_path, subjects=subject_names)

    #    df_acts = [data[f'df_activities_{sub}'] for sub in subject_names]
    #    df_devs = data['df_devices']

    #    states = (dev_type_trigger == 'state')
    #    fig_and = activities_and_devices(df_acts[0], df_devs, states=states)
    #    return fig_and





def _is_trigger(ctx, val):
    return ctx.triggered[0]['prop_id'].split('.')[0] == val


def _get_trigger_value(ctx):
    return ctx.triggered[0]['value']



from pathlib import Path

import pandas as pd
from pyadlml.dataset._core.obj import Data
from pyadlml.constants import ACTIVITY, VALUE, \
    START_TIME, END_TIME, TIME, NAME, DEVICE
from pyadlml.dataset._core.activities import correct_activities
from pyadlml.dataset._core.devices import correct_devices, CORRECTION_TS, CORRECTION_ONOFF_INCONS
from pyadlml.dataset.io import fetcher, correct_acts_and_devs, download_from_mega
import numpy as np
from pyadlml.dataset._core.devices import device_boolean_on_states_to_events
import os

"""
Transferring Knowledge of Activity Recognition across Sensor Networks
T.L.M. van Kasteren, G. Englebienne and B.J.A. Kr�se
In Proceedings of the Eighth International Conference on Pervasive Computing (Pervasive 2010). Helsinki, Finland, 2010. 

"""


KASTEREN_2010_URL = 'https://mega.nz/file/VIhnxAxB#3UI77ZA1uh0tRiT6vHTfhYolm-uxbAXuV2TxxIyQ2AU'
KASTEREN_2010_FILENAME = 'kasteren_2010.zip'

@correct_acts_and_devs
@fetcher('kasteren_2010', download_from_mega, KASTEREN_2010_FILENAME, KASTEREN_2010_URL)
def fetch_kasteren_2010(house='A', keep_original=False, cache=True,
                        retain_corrections=False, folder_path=None):
    """
    Fetches the amsterdam dataset from the internet. The original dataset or its cached version
    is stored in the :ref:`data home <storage>` folder.

    Parameters
    ----------
    house : str one of {'A', 'B', 'C'}, default='A'
        The house to load from. Every house represents a different experiment at a
        different time.
    keep_original : bool, default=False
        Determines whether the original dataset is deleted after downloading
        or kept on the hard drive.
    cache : bool, default=True
        Determines whether the data object should be stored as a binary file for quicker access.
        For more information how caching is used refer to the :ref:`user guide <storage>`.
    retain_corrections : bool, default=False
        When set to *true*, data points that change or drop during preprocessing
        are listed in respective attributes of the data object. Fore more information
        about error correction refer to the :ref:`user guide <error_correction>`.
    folder_path : str, default=None
        If set the dataset is loaded from the specified folder.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_kasteren
    >>> data = fetch_kasteren_2010()
    >>> dir(data)
    >>> [..., df_activities, df_devices, ...]

    Returns
    -------
    data : object
    """
    assert house in ['A', 'B', 'C']

    # load activity structure
    path = Path(folder_path).joinpath('datasets', 'house%s'%(house))
    df_act = _load_activities(path, house)
    df_dev = _load_devices(path, house)

    lst_act = df_act[ACTIVITY].unique()
    lst_dev = df_dev[DEVICE].unique()

    data = Data(df_act, df_dev,
                activity_list=lst_act,
                device_list=lst_dev
    )

    return data



def _load_activities(path, house):

    from scipy.io import loadmat
    tmp = loadmat(str(path.joinpath(f'actStructHouse{house}.mat')))
    acts = tmp['activityStructure'][0][0][0]

    # load act and devs
    tmp = loadmat(str(path.joinpath(f'senseandactLabelsHouse{house}.mat')))

    # load activity labels
    act_map = {}
    act_labels = tmp['activity_labels']
    nr_acts = act_labels.shape[0]
    for i in range(nr_acts):
        try:
            # (i+1) because matlab index is from 1 to n
            act_map[i+1] = act_labels[i][0][0]
        except IndexError:
            pass
    df_act = pd.DataFrame({START_TIME: acts[:,0], END_TIME: acts[:,1], ACTIVITY: acts[:,2]})
    df_act[START_TIME] = df_act[START_TIME].apply(matlab_date_to_timestamp)
    df_act[END_TIME] = df_act[END_TIME].apply(matlab_date_to_timestamp)
    df_act[ACTIVITY] = df_act[ACTIVITY].map(act_map)

    if house == 'B':
        # the activity 73 in house c maps to a non-existing activity (16.0)
        #                   start_time                   end_time  activity
        #73 2009-07-29 10:52:57.898994 2009-07-29 10:56:45.394994      16.0
        df_act = df_act.dropna()

    # Capitalize every first letter of the activity
    df_act[ACTIVITY] = df_act[ACTIVITY].apply(lambda x: x[0].upper() + x[1:])

    return df_act.sort_values(by=START_TIME).reset_index(drop=True)




def _load_devices(path, house):

    from scipy.io import loadmat
    # load device structure
    tmp =loadmat(str(path.joinpath('sensorStructHouse%s.mat'%(house))))
    devs = tmp['sensorStructure'][0][0][0]

    # load act and devs
    tmp =loadmat(str(path.joinpath('senseandactLabelsHouse%s.mat'%(house))))

    # load device labels
    if house == 'A':
        dev_map = {}
        dev_labels = tmp['sensor_labels']
        nr_devs = dev_labels.shape[0]
        for i in range(nr_devs):
            dev_map[dev_labels[i][0][0][0]] = dev_labels[i][1][0]
    elif house == 'B':
        dev_map = _dev_map_House_B()
    elif house == 'C':
        dev_map = _dev_map_House_C()

    df_dev = pd.DataFrame({START_TIME: devs[:, 0], END_TIME: devs[:, 1], DEVICE: devs[:, 2]})
    df_dev[START_TIME] = df_dev[START_TIME].apply(matlab_date_to_timestamp)
    df_dev[END_TIME] = df_dev[END_TIME].apply(matlab_date_to_timestamp)
    df_dev[DEVICE] = df_dev[DEVICE].map(dev_map)

    if house == 'B':
        # contains a device (23) where no corresponding label exists
        # therefore those two entries are dropped
        #                     start_time                   end_time  device
        #2003 2009-07-24 16:16:20.999997 2009-07-24 16:16:22.000001    23.0
        #2005 2009-07-24 16:22:09.000000 2009-07-24 16:22:10.999999    23.0
        df_dev = df_dev.dropna()

    df_dev = device_boolean_on_states_to_events(df_dev)
    return df_dev

from datetime import datetime, timedelta
def matlab_date_to_timestamp(matlab_datenum):
    """ converts a matlab datenum representation to pandas timestamp"""
    # convert matlab dateformat to python
    # copied from @actstruct/actstruct.m
    #     Usage Example: as = actstruct(732877.4520, 732877.4521, 20)
    #     activity 20 took place on 19-Juli-2006 from 10:50:52 till 10:51:01
    # https: // stackoverflow.com / questions / 13965740 / converting - matlabs - datenum - format - to - python
    # TODO the solution has not enough precision try other methods that are
    return pd.Timestamp(datetime.fromordinal(int(matlab_datenum))
                        + timedelta(days=matlab_datenum%1)
                        - timedelta(days=366))

assert pd.Timestamp('19-Jul-2006 10:50:52.800004') == matlab_date_to_timestamp(732877.4520)
assert pd.Timestamp('19-Jul-2006 10:51:01.44') == matlab_date_to_timestamp(732877.4521)


def _dev_map_House_C():
    """ translates dutch device names to english and drop the device type"""
    return {5: 'pressure mat bed right',        # mat bed rechts, drukmat
            6: 'pressure mat couch',            # 'mat bank, huiskamer'
            7: 'freezer reed',                  # 'vriezer, reed'
            8: 'toilet flush upstairs',         # 'toilet flush boven, flush ',
            10: 'toilet flush downstairs',      # 'toilet flush beneden. flush '
            11: 'bathroom door left',           # 'badkamer klapdeur links'
            13: 'cutlery kwik',                 # 'zelfde la als 18, kwik'
            15: 'keys',                         # 'La sleutels'
            16: 'bathroom door left',           # 'badkamer klapdeur links',
            18: 'cutlery kwik',                 # 'bestek la, kwik sensor '
            20: 'pans cupboard reed',           # 'kastje pannen, reed '
            21: 'microwave reed',               # 'magnetron, reed '
            22: 'cupboard leftovers reed',      # 'kastje restjesbakjes, reed '
            23: 'cabinet plates spices reed',   # 'kastje borden/kruiden,reed '
            25: 'toilet door downstairs',       # 'deur toilet beneden',
            27: 'cabinet cups/bowl/tuna reed',  #'astje cups/bowl/tuna, reed ',
            28: 'front door reed',              #'voordeur, reed ',
            29: 'door bedroom',                 #'deur slaapkamer',
            30: 'refrigerator',                 #'koelkast, reed ',
            35: 'bathtub pir',                  #'badkuip, pir ',
            36: 'dresser pir',
            38: 'washbasin flush upstairs',     #'wasbak boven, flush ',
            39: 'pressure mat bed left',        #'mat bed links, drukmat '
            }

def _dev_map_House_B():
    """ translates dutch device names to english and drop the device type"""
    return {1: 'toilet door',
            2: 'toaster *dead*',
            3: 'fridge',
            4: 'dead',
            5: 'cupboard groceries',
            6: 'toilet flush',
            7: 'frontdoor',
            8: 'dead',
            9: 'cupboard plates',
            10: 'Bedroom door',
            11: 'temp geyser',                  #'temp geiser',
            12: 'press bed right',              #'press bed  rechts',
            13: 'press bed left',               #'press bed links',
            14: 'cutlery drawer kwick',         #'kwik cutlary drawer',
            15: 'kwik stove lid',
            16: 'bedroom pir',                  #'PIR slaapkamer',
            17: 'temp shower',
            18: 'kwik dresser',
            19: 'bathroom pir',                 #'PIR badkamer',
            20: 'pressure mat office chair',    #'Drukmat bureaustoel (piano)',
            21: 'sink float',                   #'gootsteen float',
            22: 'pressure mat server corner',   #'drukmat stoel serverhoekje',
            24: 'balcony door',                 #'balkon deur',
            25: 'frame',                        #'raam',
            26: 'toaster',
            27: 'microwave',                    #'magnetron',
            28: 'kitchen pir',                  #'PIR keuken'
    }